#!/usr/bin/env python
# coding: utf-8

"""Defines the VIS image_records table using the SQLAlchemy ORM."""

# Copyright 2022-2024, United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

from datetime import datetime, timedelta, timezone
import enum
from warnings import warn
import xml.etree.ElementTree as ET

from sqlalchemy import (
    Boolean,
    DateTime,
    Enum,
    Float,
    ForeignKey,
    Identity,
    Integer,
    String,
)
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import mapped_column, relationship, synonym, validates

from vipersci.pds import Purpose
from vipersci.pds.pid import VISID, vis_instruments
from vipersci.pds.xml import find_text, ns
from vipersci.pds.datetime import fromisozformat, isozformat
from vipersci.vis.header import pga_gain as header_pga_gain
from vipersci.vis.db import Base
import vipersci.vis.db.validators as vld


class ImageType(enum.Flag):
    """This Flag class can be used to interpret the outputImageMask but not the
    immediateDownloadInfo Yamcs parameters, because only a single flag value can
    be set."""

    LOSSLESS_ICER_IMAGE = 1
    # RESERVED_2 = 2
    # RESERVED_4 = 4
    LOSSY_ICER_IMAGE = 8
    SLOG_ICER_IMAGE = 16

    @classmethod
    def _missing_(cls, value):
        return None


class ProcessingStage(enum.Flag):
    # PROCESS_RESERVED = 1
    FLATFIELD = 2
    # PROCESS_RESERVED_2 = 4
    LINEARIZATION = 8
    SLOG = 16


class ImageRecord(Base):
    """An object to represent rows in the image_records table for VIS."""

    # This class is derived from SQLAlchemy's orm.DeclarativeBase
    # which means that it has a variety of class properties that are
    # then swept up into properties on the instantiated object via
    # super().__init__().

    # The table represents many of these objects, so the __tablename__ is
    # plural while the class name is singular.
    __tablename__ = "image_records"

    # The mapped_column() names below should use "snake_case" for the names that are
    # committed to the database as column names.  Furthermore, those names
    # should be similar, if not identical, to the PDS4 Class and Attribute
    # names that they represent.  Other names (like Yamcs parameter camelCase
    # names) are implemented as synonyms. Aside from the leading "id" column,
    # the remainder are in alphabetical order, since there are so many.

    id = mapped_column(Integer, Identity(start=1), primary_key=True)
    adc_gain = mapped_column(
        Integer, nullable=False, doc="ADC_GAIN from the MCSE Image Header."
    )
    adcGain = synonym("adc_gain")
    auto_exposure = mapped_column(
        Boolean,
        nullable=False,
        doc="AUTO_EXPOSURE from the MCSE Image Header.",
    )
    autoExposure = synonym("auto_exposure")
    # bad_pixel_table_id = mapped_column(
    #     Integer,
    #     nullable=False,
    #     # There is a Defective Pixel Map (really a list of 128 coordinates) for
    #     # each MCAM.  The "state" of this is managed by the ground and not
    #     # reflected in any Image Header information attached to an individual
    #     # image.  It is not clear how to obtain this information from Yamcs,
    #     # or even what might be recorded, so this column's value is TBD.
    # )
    capture_id = mapped_column(
        Integer,
        nullable=False,
        doc="The captureId from the command sequence.",
        # TODO: learn more about captureIds to provide better doc here.
    )
    captureId = synonym("capture_id")
    _exposure_duration = mapped_column(
        "exposure_duration",
        Integer,
        nullable=False,
        doc="The exposure time in microseconds, the result of decoding the "
        "EXP_STEP and EXP paramaters from the MCSE Image Header.",
    )
    exposureTime = synonym("exposure_duration")  # Yamcs parameter name.
    file_creation_datetime = mapped_column(
        DateTime(timezone=True),
        nullable=False,
        doc="The time at which file_name was created.",
    )
    file_md5_checksum = mapped_column(
        String,
        nullable=False,
        doc="The md5 checksum of the file described by file_path.",
    )
    file_path = mapped_column(
        String,
        nullable=False,
        doc="The absolute path (POSIX style) that contains the Array_2D_Image "
        "that this metadata refers to.",
    )
    haz1LightState = synonym("light_on_hfp")
    haz2LightState = synonym("light_on_hap")
    haz3LightState = synonym("light_on_hfs")
    haz4LightState = synonym("light_on_has")
    haz5LightState = synonym("light_on_hcp")
    haz6LightState = synonym("light_on_hcs")
    icer_byte_quota = mapped_column(
        Integer,
        doc="The byteQuota value during onboard ICER compression.  In the returned "
        "Yamcs info, the value is in kilobytes, but this value is in bytes.",
    )
    icer_minloss = mapped_column(
        Integer, doc="The minLoss value during onboard ICER compression."
    )
    image_id = mapped_column(
        Integer,
        nullable=False,
        doc="The IMG_ID from the MCSE Image Header used for CCU storage and "
        "retrieval.",
    )
    image_nickname = mapped_column(
        String,
        nullable=True,
        doc="This was designed to be a unique nickname for the waypoint+camera. It has "
        "the form: <camera>-<waypoint>-<index>, where the index is represented as "
        "letters A, B, C, ..., Z, AA, AB, ...",
    )
    # This image_request_id column and image_request relationship allow a many
    # ImageRecords to one ImageRequest relationship, and the nullable allows it to be
    # optional.  So an ImageRecord may be connected to an ImageRequest, but it may not.
    image_request_id = mapped_column(ForeignKey("image_requests.id"), nullable=True)
    image_request = relationship("ImageRequest", back_populates="image_records")

    # The image_tags and image_tag_associations allow a many ImageTag to many
    # ImageRecord relationship.
    image_tags = relationship(
        "ImageTag",
        secondary="junc_image_record_tags",
        back_populates="image_records",
        viewonly=True,
    )
    image_tag_associations = relationship(
        "JuncImageRecordTag", back_populates="image_record"
    )
    imageHeight = synonym("lines")
    imageId = synonym("image_id")
    imageNickname = synonym("image_nickname")
    imageWidth = synonym("samples")
    instrument_name = mapped_column(
        String, nullable=False, doc="The full name of the instrument."
    )
    instrument_temperature = mapped_column(
        Float,
        nullable=False,
        doc="The TEMPERATURE from the MCSE Image Header.  TBD how to convert "
        "this 16-bit integer into degrees C.",
    )
    # There is a sensor in the camera body (PT1000) which is apparently not
    # connected (sigh).  And there is also a sensor external to each camera
    # body (AD590), need to track down its Yamcs feed.
    light_on_hfp = mapped_column(
        Boolean,
        nullable=True,
        doc="haz1LightState as reported by image metadata.",
    )
    light_on_hap = mapped_column(
        Boolean,
        nullable=True,
        doc="haz2LightState as reported by image metadata.",
    )
    light_on_hfs = mapped_column(
        Boolean,
        nullable=True,
        doc="haz3LightState as reported by image metadata.",
    )
    light_on_has = mapped_column(
        Boolean,
        nullable=True,
        doc="haz4LightState as reported by image metadata.",
    )
    light_on_hcp = mapped_column(
        Boolean,
        nullable=True,
        doc="haz5LightState as reported by image metadata.",
    )
    light_on_hcs = mapped_column(
        Boolean,
        nullable=True,
        doc="haz6LightState as reported by image metadata.",
    )
    light_on_nl = mapped_column(
        Boolean,
        nullable=True,
        doc="navLeftLightState as reported by image metadata.",
    )
    light_on_nr = mapped_column(
        Boolean,
        nullable=True,
        doc="navRightLightState as reported by image metadata.",
    )
    lines = mapped_column(
        Integer,
        nullable=False,
        doc="The imageHeight parameter from the Yamcs imageHeader.",
    )
    _lobt = mapped_column(
        "lobt",
        Integer,
        nullable=False,
        doc="The TIME_TAG from the MCSE Image Header.",
    )
    navLeftLightState = synonym("light_on_nl")
    navRightLightState = synonym("light_on_nr")
    # mcam_id, The MCAM_ID from the MCSE Image Header is not returned to the ground.
    offset = mapped_column(
        Integer,
        nullable=False,
        doc="The OFFSET parameter from the MCSE Image Header describing the dark "
        "level offset.",
    )
    # onboard_compression_ratio = mapped_column(
    #     Float,
    #     doc="The PDS img:Onboard_Compression parameter.  Will be NULL for "
    #     "uncompressed images, 1 for LOSSLESS compressed, 16 for 16x compressed, etc."
    #     # This is a PDS img:Onboard_Compression parameter which is the ratio
    #     # of the size, in bytes, of the original uncompressed data object
    #     # to its compressed size.  This operation is done by RFSW, and fixed
    #     # by that process.  The output_image_mask determines what happened,
    #     # and then we must look up to find this value.
    # )
    output_image_mask = mapped_column(
        Integer,
        nullable=False,
        doc="The outputImageMask from the Yamcs imageHeader.  For each downlinked "
        "image this can be exactly one value of the ImageType class.  This value "
        "indicates whether the image is SLoG or not, and what level of compression "
        "has been set.",
    )
    # outputImageType is redundant with the outputImageMask, as it is just the
    # longform name of the value specified in the outputImageMask.
    outputImageMask = synonym("output_image_mask")
    # The pano_records and pano_record_associations allow a many PanoRecords to many
    # ImageRecords relationship.
    pano_records = relationship(
        "PanoRecord",
        secondary="junc_image_pano",
        back_populates="image_records",
        viewonly=True,
    )
    pano_record_associations = relationship(
        "JuncImagePano", back_populates="image_record"
    )
    _pid = mapped_column(
        "product_id", String, nullable=False, unique=True, doc="The PDS Product ID."
    )
    padding = mapped_column(
        Integer,
        nullable=False,
        doc="The padding parameter from the Yamcs imageHeader.",
        # Not sure what this value means or where it comes from.
    )
    pga_gain = mapped_column(
        Float,
        nullable=False,
        doc="The translated floating point multiplier derived from PGA_GAIN "
        "from the MCSE Image Header.",
    )
    pgaGain = synonym("pga_gain")
    processing_info = mapped_column(
        Integer,
        nullable=False,
        doc="The processingInfo parameter from the Yamcs imageHeader. This integer "
        "value must correspond to a valid value of ProcessingStage, and indicates "
        "what onboard processing occurred.",
    )
    processingInfo = synonym("processing_info")
    requestId = synonym("image_request_id")
    samples = mapped_column(
        Integer,
        nullable=False,
        doc="The imageWidth parameter from the Yamcs imageHeader.",
    )
    # slog = mapped_column(
    #     Boolean,
    #     nullable=False,
    #     doc="Indicates whether onboard SLoG processing occurred.",
    # )
    software_name = mapped_column(String, nullable=False)
    software_version = mapped_column(String, nullable=False)
    software_program_name = mapped_column(String, nullable=False)
    start_time = mapped_column(DateTime(timezone=True), nullable=False)
    stereo = mapped_column(
        Boolean,
        nullable=False,
        doc="The stereo parameter from the Yamcs imageHeader.",
        # TODO: learn more about stereo to provide better doc here.
    )
    stop_time = mapped_column(DateTime(timezone=True), nullable=False)
    temperature = synonym("instrument_temperature")
    unique_capture_id = mapped_column(
        Integer,
        nullable=True,
        doc="The unique portion from the lower 16 bits of the captureId.  Ironically, "
        "this isn't guaranteed to be a mission-long globally unique value, only "
        "locally unique in time.  It gets set and linearly increases, but can "
        "be reset to zero.",
    )
    verification_notes = mapped_column(
        String,
        nullable=True,
        doc="Any notes about the verification of this image by the VIS Operator.",
    )
    verification_purpose = mapped_column(
        Enum(Purpose), nullable=True, doc="Purpose of Observation, as defined by PDS."
    )
    verified = mapped_column(
        Boolean,
        nullable=True,
        doc="True if a VIS Operator has indicated that this image is a good image, if "
        "false, the VIS Operator has determined that there is an error of some kind "
        "with this image.",
    )
    verifier = mapped_column(
        String,
        nullable=True,
        doc="The name of the individual that reviewed this image.",
    )
    voltage_ramp = mapped_column(
        Integer,
        nullable=False,
        doc="The VOLTAGE_RAMP parameter from the MCSE Image Header.",
    )
    voltageRamp = synonym("voltage_ramp")
    waypoint_id = mapped_column(
        Integer,
        nullable=True,
        doc="The waypoint id extracted from the upper 16 bits of the captureId.",
    )
    yamcs_generation_time = mapped_column(
        DateTime(timezone=True),
        nullable=False,
        doc="The generation time of the source record from Yamcs.",
    )
    yamcs_name = mapped_column(
        String,
        nullable=False,
        doc="The full parameter name from Yamcs that this product data came from, "
        "formatted like a / separated string.",
    )
    yamcs_reception_time = mapped_column(
        DateTime(timezone=True),
        nullable=True,
        doc="The reception time of the source record from Yamcs.",
    )

    def __init__(self, **kwargs):
        if "lobt" in kwargs:
            lobt_dt = datetime.fromtimestamp(kwargs["lobt"], tz=timezone.utc)

        if kwargs.keys() >= {"start_time", "lobt"}:
            if isinstance(kwargs["start_time"], str):
                dt = fromisozformat(kwargs["start_time"])
            else:
                dt = kwargs["start_time"]

            if lobt_dt != dt:
                raise ValueError(
                    f"The start_time {kwargs['start_time']} does not equal "
                    f"the lobt ({kwargs['lobt']}, {lobt_dt})"
                )

        # Exposure duration is a hybrid_property that also sets the stop_time,
        # if super().__init() processes exposure duration while self.start_time
        # is still None, then object initiation will fail.  Removing it from
        # the parameters we pass to super().__init() and then setting it
        # after avoids this error condition.
        exp_dur = None
        for k in ("exposureTime", "exposure_duration"):
            if k in kwargs:
                exp_dur = kwargs[k]
                del kwargs[k]

        # If present, product_id needs some special handling:
        if "product_id" in kwargs:
            pid = VISID(kwargs["product_id"])
            del kwargs["product_id"]
        else:
            pid = False

        # Adjust the byteQuota value
        if "byteQuota" in kwargs and "icer_byte_quota" not in kwargs:
            kwargs["icer_byte_quota"] = int(kwargs["byteQuota"]) * 1000
            del kwargs["byteQuota"]

        if "minLoss" in kwargs and "icer_minloss" not in kwargs:
            kwargs["icer_minloss"] = int(kwargs["minLoss"])
            del kwargs["minLoss"]

        rpargs = dict()
        otherargs = dict()
        for k, v in kwargs.items():
            if k in self.__table__.columns or k in self.__mapper__.synonyms:
                rpargs[k] = v
            else:
                otherargs[k] = v

        # Instantiate early, so that the parent orm_declarative Base can
        # resolve all of the synonyms.
        super().__init__(**rpargs)

        # Ensure stop_time consistency by setting this *after* start_time is set in
        # super().__init__()
        self.exposure_duration = exp_dur

        # Ensure instrument_name consistency and existence.
        if "instrument_name" in kwargs:
            self.instrument_name = VISID.instrument_name(self.instrument_name)
        elif "yamcs_name" in kwargs:
            self.instrument_name = VISID.instrument_name(
                self.yamcs_name.split("/")[-1].replace("_", " ")
            )

        if "cameraId" in otherargs:
            if VISID.instrument_name(otherargs["cameraId"]) != self.instrument_name:
                warn(
                    f"cameraId ({otherargs['cameraId']}) does not match the "
                    f"instrument_name ({self.instrument_name})."
                )

        # # Derive slog, if possible.
        # if self.processing_info is not None:
        #     try:
        #         if "slog" in kwargs:
        #             # check against processing_info
        #             if kwargs[
        #                 "slog"
        #             ] and ProcessingStage.PROCESS_SLOG not in ProcessingStage(
        #                 self.processing_info
        #             ):
        #                 warn(
        #                     "slog is True, but ProcessingStage.PROCESS_SLOG not "
        #                     f"in {ProcessingStage(self.processing_info)}"
        #                 )

        #             if not kwargs[
        #                 "slog"
        #             ] and ProcessingStage.PROCESS_SLOG in ProcessingStage(
        #                 self.processing_info
        #             ):
        #                 warn(
        #                     "slog is False, but ProcessingStage.PROCESS_SLOG "
        #                     f"in {ProcessingStage(self.processing_info)}"
        #                 )
        #         else:
        #             if ProcessingStage.PROCESS_SLOG in ProcessingStage(
        #                 self.processing_info
        #             ):
        #                 self.slog = True
        #             else:
        #                 self.slog = False
        #     except ValueError as err:
        #         warn(str(err))
        #         if "slog" not in kwargs and "yamcs_name" in kwargs:
        #             self.slog = self.yamcs_name.endswith("slog")
        # else:
        #     if "slog" not in kwargs and "yamcs_name" in kwargs:
        #         self.slog = self.yamcs_name.endswith("slog")

        # Ensure product_id consistency
        if pid:
            # Check datetimes
            if "lobt" in kwargs:
                if pid.datetime() != lobt_dt:
                    raise ValueError(
                        f"The product_id datetime ({pid.datetime()}) and the "
                        f"provided lobt ({kwargs['lobt']}, {lobt_dt}) disagree."
                    )

            if "start_time" in kwargs and pid.datetime() != self.start_time:
                raise ValueError(
                    f"The product_id datetime ({pid.datetime()}) and the "
                    f"provided start_time ({kwargs['start_time']}) disagree."
                )

            # Check instrument
            if (
                self.instrument_name is not None
                and vis_instruments[pid.instrument] != self.instrument_name
            ):
                raise ValueError(
                    f"The product_id instrument code ({pid.instrument}) and "
                    f"the provided instrument_name "
                    f"({self.instrument_name}) disagree."
                )

            # Check compression letter
            if self.output_image_mask is None:
                if self.yamcs_name is not None:
                    if "slog" in self.yamcs_name and pid.compression != "s":
                        raise ValueError(
                            f"The product_id compression code ({pid.compression}) is "
                            "not s, but yamcs_name indicates it should be "
                            f"({self.yamcs_name}). "
                        )
            else:
                t = ImageType(self.output_image_mask)
                if ImageType.SLOG_ICER_IMAGE == t and pid.compression == "s":
                    pass
                elif (
                    VISID.compression_letter(compression_ratio(self.icer_byte_quota))
                    != pid.compression
                ):
                    raise ValueError(
                        f"The product_id compression code ({pid.compression}) and "
                        "the compression ratio "
                        f"({compression_ratio(self.icer_byte_quota)}) based on "
                        f"the icer_byte_quota ({self.icer_byte_quota}) disagree."
                    )
        elif self.start_time is not None and self.instrument_name is not None:
            c = None
            if self.output_image_mask is not None:
                try:
                    if ImageType(self.output_image_mask) == ImageType.SLOG_ICER_IMAGE:
                        c = "s"
                except ValueError:
                    # output_image_mask has bad value
                    pass

            if c is None and self.yamcs_name is not None and "slog" in self.yamcs_name:
                c = "s"

            if c is None and self.icer_byte_quota is not None:
                c = compression_ratio(self.icer_byte_quota)

            if c is None:
                raise ValueError(
                    "Could not determine the compression information "
                    f"from output_image_mask ({self.output_image_mask}), "
                    f"processing_info ({self.processing_info}), or "
                    f"icer_byte_quota ({self.icer_byte_quota})."
                )

            pid = VISID(
                self.start_time.date(),
                self.start_time.time(),
                self.instrument_name,
                c,
            )
        else:
            got = dict()
            for k in (
                "product_id",
                "start_time",
                "instrument_name",
                "output_image_mask",
                "processing_info",
                "icer_byte_quota",
            ):
                v = getattr(self, k)
                if v is not None:
                    got[k] = v

            raise ValueError(
                "Either product_id must be given, or each of start_time, "
                f"instrument_name, and output_image_mask plus some other things."
                f"Got: {got}"
            )

        self._pid = str(pid)

        if self.capture_id is not None and self.capture_id > int(
            "1111111111111111", base=2
        ):  # 65535
            self.waypoint_id = int(bin(self.capture_id)[:-16], base=2)
            self.unique_capture_id = int(bin(self.capture_id)[-16:], base=2)

        # Is this really a good idea?  Not sure.  This instance variable plus
        # label_dict() and update() allow other key/value pairs to be carried around
        # in this object, which is handy.  If these are well enough known, perhaps
        # they should just be pre-defined properties and not left to chance?
        self.labelmeta = otherargs

        return

    def __lt__(self, other):
        if isinstance(other, self.__class__):
            return VISID(self.product_id) < VISID(other.product_id)
        else:
            return NotImplemented

    @hybrid_property
    def exposure_duration(self):
        return self._exposure_duration

    @exposure_duration.inplace.setter
    def _exposure_duration_setter(self, value: int):
        """Takes an exposure time in microseconds."""
        self._exposure_duration = value
        self.stop_time = self.start_time + timedelta(microseconds=value)

    @hybrid_property
    def lobt(self):
        return self._lobt

    @lobt.inplace.setter
    def _lobt_setter(self, lobt):
        self._lobt = lobt
        self.start_time = datetime.fromtimestamp(lobt, tz=timezone.utc)

    @hybrid_property
    def product_id(self):
        # Really am going back and forth about whether this should be returned as
        # a full VISID object or just as the string as it is now.
        return self._pid

    @product_id.inplace.setter
    def _product_id_setter(self, pid):
        # In this class, the source of product_id information really is what
        # comes from Yamcs, and so this should not be monkeyed with.  Theoretically
        # changing this would imply changes to start time, lobt, stop time,
        # intrument name and onboard_compression_ratio directly, but those changes then
        # also divorce this object from the Yamcs parameters that it came from and
        # has all manner of other implications.  So at this time, this can only be
        # set when this object is instantiated.
        raise NotImplementedError(
            "product_id cannot be set directly after instantiation."
        )

    @validates("pga_gain")
    def validate_pga_gain(self, key, value):
        return header_pga_gain(value)

    @validates(
        "file_creation_datetime",
        "start_time",
        "stop_time",
        "yamcs_generation_time",
        "yamcs_reception_time",
    )
    def validate_datetime_asutc(self, key, value):
        return vld.validate_datetime_asutc(key, value)

    @validates(
        "light_on_hap",
        "light_on_has",
        "light_on_hcp",
        "light_on_hcs",
        "light_on_hfp",
        "light_on_hfs",
        "light_on_nl",
        "light_on_nr",
    )
    def validate_lights(self, key, value):
        if isinstance(value, str):
            if value.casefold() == "on":
                return True
            elif value.casefold() == "off":
                return False

        return bool(value)

    @validates("output_image_mask")
    def validate_output_image_mask(self, key, value):
        try:
            ImageType(value)
        except ValueError:
            warn(f"{key} ({value}) is not one of {list(ImageType)}")

        return value

    @validates("processing_info")
    def validate_processing_info(self, key, value):
        try:
            ProcessingStage(value)
        except ValueError:
            warn(f"{key} ({value}) is not one of {list(ProcessingStage)}")

        return value

    def asdict(self):
        d = {}

        for c in self.__table__.columns:
            if isinstance(getattr(self, c.name), datetime):
                d[c.name] = isozformat(getattr(self, c.name))
            else:
                d[c.name] = getattr(self, c.name)

        if hasattr(self, "labelmeta"):
            d.update(self.labelmeta)

        return d

    @classmethod
    def from_xml(cls, text: str):
        """
        Returns an instantiated RawProduct object from parsing the provided *text*
        as XML.
        """
        d = {}

        root = ET.fromstring(text)
        lid = find_text(root, "./pds:Identification_Area/pds:logical_identifier").split(
            ":"
        )

        if lid[3] != "viper_vis":
            raise ValueError(
                f"XML text has a logical_identifier which is not viper_vis: {lid[3]}"
            )

        if lid[4] != "raw":
            raise ValueError(
                f"XML text has a logical_identifier which is not raw: {lid[4]}"
            )
        d["product_id"] = lid[5]

        d["auto_exposure"] = (
            True if find_text(root, ".//img:exposure_type") == "Auto" else False
        )
        d["bad_pixel_table_id"] = int(
            find_text(root, ".//img:bad_pixel_replacement_table_id")
        )
        d["exposure_duration"] = int(
            find_text(root, ".//img:exposure_duration", unit_check="microseconds")
        )

        d["file_creation_datetime"] = fromisozformat(
            find_text(root, ".//pds:creation_date_time")
        )
        d["file_path"] = find_text(root, ".//pds:file_name")

        # for k, v in luminaire_names.items():
        #     light = root.find(f".//img:LED_Illumination_Source[img:name='{k}']", ns)
        #     d[v] = (
        #         True if _find_text(light, "img:illumination_state") == "On" else False
        #     )

        osc = root.find(".//pds:Observing_System_Component[pds:type='Instrument']", ns)
        d["instrument_name"] = find_text(osc, "pds:name")

        d["instrument_temperature"] = float(
            find_text(root, ".//img:temperature_value", unit_check="K")
        )

        aa = root.find(".//pds:Axis_Array[pds:axis_name='Line']", ns)
        d["lines"] = int(find_text(aa, "./pds:elements"))
        d["file_md5_checksum"] = find_text(root, ".//pds:md5_checksum")
        d["mission_phase"] = find_text(root, ".//msn:mission_phase_name")
        d["offset"] = find_text(root, ".//img:analog_offset")

        try:
            d["onboard_compression_ratio"] = float(
                find_text(root, ".//img:onboard_compression_ratio")
            )
        except ValueError:
            pass

        d["purpose"] = find_text(root, ".//pds:purpose")

        aa = root.find(".//pds:Axis_Array[pds:axis_name='Sample']", ns)
        d["samples"] = int(find_text(aa, "./pds:elements"))

        sw = root.find(".//proc:Software", ns)
        d["software_name"] = find_text(sw, "./proc:name")
        d["software_version"] = find_text(sw, "./proc:software_version_id")
        d["software_program_name"] = find_text(sw, "./proc:Software_Program/proc:name")

        # Start times must be on the whole second, which is why we don't use
        # fromisozformat() here.
        d["start_time"] = datetime.strptime(
            find_text(root, ".//pds:start_date_time"), "%Y-%m-%dT%H:%M:%SZ"
        ).replace(tzinfo=timezone.utc)

        d["stop_time"] = fromisozformat(find_text(root, ".//pds:stop_date_time"))

        return cls(**d)

    def update(self, other):
        for k, v in other.items():
            if k in self.__table__.columns or k in self.__mapper__.synonyms:
                setattr(self, k, v)
            else:
                self.labelmeta[k] = v


def compression_ratio(byte_quota):
    """Returns the result of dividing the number of bytes in a grayscale image
    (2048 * 2048 * 2 == 8,388,608) by the byte_quota of the returned image.
    """
    return (2048 * 2048 * 2) / byte_quota



"""NSS model module.

This module contains functions for ingesting the text file format that the
NSS team produces and for providing interpolation for the forward and reverse
models.
"""

# Copyright 2022, United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

import logging
from typing import Any, Iterable, Sequence, Union
from os import PathLike

import numpy as np
import numpy.typing as npt
from scipy.interpolate import RegularGridInterpolator

logger = logging.getLogger(__name__)

# type alias for the exhaustive variety of arguments that np.genfromtxt() takes.
Readable = Union[str, PathLike, Iterable[str], Iterable[bytes]]


class DataSimulator:
    """
    The DataSimulator object is initiated with inverse model model files,
    and then can be called to return simulated detector
    data given burial depth and water equivalent hydrogen values.
    """

    def __init__(
        self,
        det1: Readable,
        det2: Readable,
        bounds_error: bool = True,
        fill_value: Any = np.nan,
        rng: np.random.Generator = np.random.default_rng(),
    ):
        """
        :param det1: The path to the Detector 1 inverse model CSV file.
        :type det1: Path
        :param det2: The path to the Detector 2 inverse model CSV file.
        :type det2: Path
        :param bounds_error: If True, when interpolated values are requested
        outside of the domain of the inverse models, a ValueError is raised.
        If False, then fill_value is used.
        :param fill_value: If provided, the value to use for points outside
        of the interpolation domain. If None, values outside the domain are
        extrapolated.
        :param rng: A random generator for the object to use when Poisson
        noise is requested.

        The *bounds_error* and *fill_value* arguments are passed on to the
        model() function, please see its documentation for more information.
        """
        self.det1_model = model(det1, bounds_error=bounds_error, fill_value=fill_value)
        self.det2_model = model(det2, bounds_error=bounds_error, fill_value=fill_value)
        self.rng = rng
        return

    def __call__(
        self,
        bd: Union[float, Sequence, np.ndarray],
        weh: Union[float, Sequence, np.ndarray],
        poisson: bool = False,
    ):
        """
        Returns simulated detector 1 and detector 2 values.

        :param bd:  A single value or sequence of burial depth.
        :param weh: A single value or sequence of water equivalent hydrogen.
        :param poisson: A boolean (default False) indicating whether Poisson
        noise should be added to the returned simulated data.
        :returns: A two-tuple of values or np.arrays.  If *bd* and *weh* are
        singular elements, then a two-tuple is returned of the detector 1 and
        detector 2 values at that location.  If *bd* and *weh* contain more than
        one value each, then the returned two-tuple will be a numpy array
        of detector 1 values, and a numpy array of detector 2 values.
        """
        is_arraylike = True if hasattr(bd, "__iter__") else False

        d1 = self.det1_model(np.column_stack((bd, weh)))
        d2 = self.det2_model(np.column_stack((bd, weh)))

        if poisson:
            d1, d2 = self.rng.poisson(lam=(d1, d2))

        if not is_arraylike:
            return d1.item(), d2.item()

        return d1, d2


class DataModeler:
    """
    The DataModeler object is initiated with model files,
    and then can be called to apply the models to a set of detector 1 & 2
    data to provide burial depth and water equivalent hydrogen values.
    """

    def __init__(
        self,
        bd_mod: Readable,
        weh_mod: Readable,
        fill_value: Any = np.nan,
    ):
        """
        :param bd_mod: The path to the BD model CSV file.
        :type bd_mod: Path
        :param weh_mod: The path to the WEH model CSV file.
        :type weh_mod: Path
        :param fill_value: fill value for nodata regions

        The *fill_value* arguments is passed on to the
        model() function. Please see its documentation for more information.
        """
        self.det1_model = model(bd_mod, bounds_error=False, fill_value=fill_value)
        self.det2_model = model(weh_mod, bounds_error=False, fill_value=fill_value)
        self.fill_value = fill_value
        return

    def __call__(
        self,
        det1: Union[float, Sequence[float], npt.ArrayLike],
        det2: Union[float, Sequence[float], npt.ArrayLike],
    ):
        """
        Returns model-provided burial depth, water equivalent hydrogen, and
        uniform water equivalent hydrogen values.

        :param det1:  A single value or sequence of detector 1 values.
        :param det2: A single value or sequence of detector 2 values.
        :returns: A three-tuple of values or np.arrays.  If *det1* and *det2* are
        singular elements, then a three-tuple is returned of the BD, WEH, and UWEH
        values at that location.  If *det1* and *det2* contain more than
        one value each, then the returned three-tuple will be a numpy array
        of BD values, a numpy array of WEH values, and a numpy array of UWEH values.
        """
        is_arraylike = True if hasattr(det1, "__iter__") else False

        if isinstance(det1, np.ma.MaskedArray) or isinstance(det2, np.ma.MaskedArray):
            d1 = det1 if isinstance(det1, np.ma.MaskedArray) else np.ma.array(det1)
            d2 = det2 if isinstance(det2, np.ma.MaskedArray) else np.ma.array(det2)

            if not np.array_equal(d1.mask, d2.mask):
                raise ValueError(
                    "The masks for the two masked arrays are not symmetric."
                )

            bd_arr = np.full_like(det1, self.fill_value, dtype=np.double)
            weh_arr = np.full_like(det2, self.fill_value, dtype=np.double)

            stacked = np.column_stack((d1.compressed(), d2.compressed()))
            bd_arr[~d1.mask] = self.det1_model(stacked)
            weh_arr[~d1.mask] = self.det2_model(stacked)
            d1_for_uweh = d1.filled(self.fill_value)

        else:
            stacked = np.column_stack((det1, det2))
            bd_arr = self.det1_model(stacked)
            weh_arr = self.det2_model(stacked)
            d1_for_uweh = det1

        uweh_arr = uniform_weh(
            d1_for_uweh, fill_value=self.fill_value, bounds_error=False
        )

        if not is_arraylike:
            return bd_arr.item(), weh_arr.item(), uweh_arr.item()

        return bd_arr, weh_arr, uweh_arr


def model(fname: Readable, **kwargs):
    """
    Returns a scipy.interpolate.RegularGridInterpolator "instance" created
    from the data provided in the NSS file *fname*.

    The returned instance can be called with an ndarray of shape (..., 2)
    which contain the x, y coordinates at which to sample the data.

    The Burial Depth and WEH files, are expected to have Detector 1 (He Cd) as
    the x-coordinate and Detector 2 (He Sn) as the y-coordinate.

    Any additional keyword arguments will be passed to
    scipy.interpolate.RegularGridInterpolator().  Please review its
    documentation for more information.
    """

    arr, row_coords, col_coords = read_csv(fname)
    # These are inverted to make calling the RGI instance more "natural" to
    # call as described in the docstring above.
    return RegularGridInterpolator(
        (col_coords, row_coords), np.transpose(arr), **kwargs
    )


def read_csv(fname: Readable):
    """
    Returns three objects: a 2D array of "data", an array of row coordinates,
    and an array of column coordinates.

    The *fname* is expected to be a CSV file provided by the NSS team which
    describes either a forward or reverse model for detector counts.  These
    files are conceptually describing a graph.  These CSV files use the first
    row and first column as "headers" to describe the data within.  The first
    element of the first row is not data and is ignored.  The
    remainder of the first row indicates the column values, and the remainder
    of the first column indicates the row values.  The remaining rectangular
    array are the data.
    """
    f = np.genfromtxt(fname, delimiter=",", missing_values="NaN")

    arr = f[1:, 1:]
    col_coords = f[0, 1:]
    row_coords = f[1:, 0]

    if col_coords[0] > col_coords[-1]:
        col_coords = np.flip(col_coords)
        arr = np.fliplr(arr)

    if row_coords[0] > row_coords[-1]:
        row_coords = np.flip(row_coords)
        arr = np.flipud(arr)

    # arr is in [row, col] order.
    return arr, row_coords, col_coords


def uniform_weh(
    measured: Union[float, Sequence[float], npt.ArrayLike],
    c0: float = 30.75,
    a: float = 0.0256,
    b: float = 1.0990,
    bounds_error: bool = True,
    fill_value: Any = np.nan,
):
    """Returns fraction of water equivalent hydrogen.

    With 1 corresponding to pure ice, and 0.01 corresponding to 1 wt% WEH.

    The value(s) of *measured* should be epithermal count rates (Detector #1),
    and this model assumes a uniformly distributed water-equivalent
    hydrogen abundance.

    The default values of c0, a, and b correspond to conditions of the Lunar
    Prospector mission in 1998/1999, during the rising phase of cycle 23.
    The state of the sun during VIPER’s mission is yet unknown but may be more
    active – this would reduce the neutron flux.

    bounds_error: bool, optional
    If True, when a value of *measured* is greater than c0 (or less than or
    equal to zero), a ValueError is raised. If False, then *fill_value* is used.

    fill_value: number, optional
    If provided, the value to use for points when *measured* triggers the
    *bounds_error*.
    """
    is_arraylike = True if hasattr(measured, "__iter__") else False
    if is_arraylike and not isinstance(measured, np.ndarray):
        m: Any = np.array(measured)
    else:
        m = measured

    # Values of *measured* larger than C0 result in unrealistic negative
    # fractions.
    if bounds_error:
        if np.any(m > c0):
            raise ValueError(f"Value(s) in measured are greater than c0, {c0}.")

        if np.any(m <= 0):
            raise ValueError("Value(s) in measured are less than or equal to zero.")

    with np.errstate(divide="ignore", invalid="ignore"):
        arr = np.float_power(a * ((c0 / m) - 1), b)

    if is_arraylike:
        arr[np.logical_or(arr < 0, arr > 1)] = fill_value
    else:
        if arr < 0 or arr > 1:
            arr = fill_value

    return arr



#!/usr/bin/env python
"""This module has tests for heatmap.generate_density_heatmap"""

# Copyright 2022, United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

import math
from typing import Tuple
import unittest

import numpy as np
import numpy.testing as npt
from rasterio import transform, windows
import shapely

from vipersci.carto import heatmap as hm


class TestDensityHeatmap(unittest.TestCase):
    def test_density_heatmap_3x3_uniform(self):
        self.uniform_density_heatmap_runner((3, 3), processes=1)

    @unittest.skip("for speedup")
    def test_density_heatmap_30x50_uniform(self):
        self.uniform_density_heatmap_runner((30, 50), processes=8)

    @unittest.skip("for speedup")
    def test_density_heatmap_31x17_uniform(self):
        self.uniform_density_heatmap_runner((31, 17), processes=8)

    @unittest.skip("for speedup")
    def test_density_heatmap_100x100_uniform(self):
        self.uniform_density_heatmap_runner((100, 100), processes=8)

    def test_density_heatmap_diag(self, processes: int = 1):
        shape = (10, 10)
        x = np.arange(0.5, shape[1] - 0.5, 1)
        y = np.arange(0.5, shape[0] - 0.5, 1)
        values = np.ones(shape[0] - 1)
        affine_transform, out_freq, out_avg, _ = hm.generate_density_heatmap(
            x, y, values, gsd=1, radius=1, padding=0, processes=processes
        )

        expected_output = np.zeros(shape)
        max_i = min(shape[0], shape[1])
        for i in range(0, max_i - 1):
            xc = i + 1
            yc = max_i - i - 1
            expected_output[xc][yc] = 1

        self.assertTrue(np.array_equal(out_avg, expected_output))
        self.assertEqual(
            affine_transform,
            transform.from_bounds(-1, 0, shape[1] - 1, shape[0], shape[1], shape[0]),
        )

        expected_freq = np.zeros(shape, dtype=np.int32)
        for i in range(1, max_i):
            xc = i
            yc = max_i - i
            expected_freq[xc][yc] = 1
        self.assertTrue(np.array_equal(expected_freq, out_freq))

    def uniform_density_heatmap_runner(
        self,
        shape: Tuple[int, int],
        gsd: float = 1,
        radius: float = 1,
        padding: float = 0,
        processes: int = 1,
    ):
        x_arr = np.arange(0.5, shape[1], 1)
        y_arr = np.arange(0.5, shape[0], 1)
        x, y = np.meshgrid(x_arr, y_arr)
        values = np.ones(len(x_arr) * len(y_arr))
        affine_transform, frequencies, out_avg, _ = hm.generate_density_heatmap(
            x.ravel(),
            y.ravel(),
            values,
            gsd=gsd,
            radius=radius,
            padding=padding,
            processes=processes,
        )

        expected_result = np.ones((shape[0] + 1, shape[1] + 1))
        for i in range(0, shape[0] + 1):
            expected_result[i][0] = 0
        for i in range(0, shape[1] + 1):
            expected_result[0][i] = 0
        self.assertTrue(np.array_equal(out_avg, expected_result))

        expected_bounds = (
            math.floor(0.5 - radius),
            math.ceil(0.5 - radius),
            math.ceil(shape[1] - 1.5 + radius),
            math.ceil(shape[0] - 0.5 + radius),
        )
        self.assertEqual(
            expected_bounds,
            windows.bounds(windows.get_data_window(out_avg), affine_transform),
        )

        expected_frequencies = np.ones((shape[0] + 1, shape[1] + 1))
        for i in range(0, shape[0] + 1):
            expected_frequencies[i][0] = 0
        for i in range(0, shape[1] + 1):
            expected_frequencies[0][i] = 0

        self.assertTrue(np.array_equal(expected_frequencies, frequencies))

    def test_sample_bounds(
        self,
        gsd: float = 1,
        radius: float = 1,
        padding: float = 0,
        processes: int = 1,
    ):
        shape = (10, 10)
        x_arr = np.arange(0.5, shape[1], 1)
        y_arr = np.arange(0.5, shape[0], 1)
        x, y = np.meshgrid(x_arr, y_arr)
        values = np.ones(len(x_arr) * len(y_arr))

        sample_bounds = shapely.geometry.Polygon(
            [(3, 3), (3, 7), (7, 7), (7, 3), (3, 3)]
        )

        affine_transform, frequencies, out_avg, _ = hm.generate_density_heatmap(
            x.ravel(),
            y.ravel(),
            values,
            gsd=gsd,
            radius=radius,
            padding=padding,
            processes=processes,
            sample_bounds=sample_bounds,
        )

        expected_result = np.ones((6, 6))
        expected_result[0][0] = 0
        expected_result[5][5] = 0
        self.assertTrue(np.array_equal(out_avg, expected_result))

        expected_bounds = (
            math.floor(3 - radius),
            math.ceil(3 - radius),
            math.ceil(7 + radius),
            math.ceil(7 + radius),
        )
        self.assertEqual(
            expected_bounds,
            windows.bounds(windows.get_data_window(out_avg), affine_transform),
        )

        self.assertTrue(np.array_equal(expected_result, frequencies))

    def test_frequency_reuse(
        self,
        gsd: float = 1,
        radius: float = 1,
        padding: float = 0,
        processes: int = 1,
    ):
        shape = (10, 10)

        x_arr = np.arange(0.5, shape[1], 1)
        y_arr = np.arange(0.5, shape[0], 1)
        x, y = np.meshgrid(x_arr, y_arr)
        values = np.ones(len(x_arr) * len(y_arr))
        (
            affine_transform,
            frequencies,
            out_avg,
            raw_frequencies,
        ) = hm.generate_density_heatmap(
            x.ravel(),
            y.ravel(),
            values,
            gsd=gsd,
            radius=radius,
            padding=padding,
            processes=processes,
        )

        affine_transform_2, frequencies_2, out_avg_2, _ = hm.generate_density_heatmap(
            x.ravel(),
            y.ravel(),
            values,
            gsd=gsd,
            radius=radius,
            padding=padding,
            processes=processes,
            frequencies=raw_frequencies,
        )

        self.assertTrue(np.array_equal(frequencies, frequencies_2))
        self.assertTrue(np.array_equal(out_avg, out_avg_2))
        self.assertEqual(affine_transform, affine_transform_2)

    def test_single_location(
        self,
        gsd: float = 1,
        radius: float = 1,
        padding: float = 0,
        processes: int = 1,
    ):
        shape = (10, 10)
        x_arr = np.ones(shape[1])
        y_arr = np.ones(shape[0])
        x, y = np.meshgrid(x_arr, y_arr)
        values = np.ones(len(x_arr) * len(y_arr))
        affine_transform, frequencies, out_avg, _ = hm.generate_density_heatmap(
            x.ravel(),
            y.ravel(),
            values,
            gsd=gsd,
            radius=radius,
            padding=padding,
            processes=processes,
        )

        expected_result = np.ones((2, 2))
        self.assertTrue(np.array_equal(out_avg, expected_result))

        expected_bounds = (
            math.floor(1 - radius),
            math.ceil(1 - radius),
            math.ceil(1 + radius),
            math.ceil(1 + radius),
        )
        self.assertEqual(
            expected_bounds,
            windows.bounds(windows.get_data_window(out_avg), affine_transform),
        )

        expected_frequencies = np.full((2, 2), shape[0] * shape[1])

        self.assertTrue(np.array_equal(expected_frequencies, frequencies))

    def test_missing_value(self):
        t, counts, avg, freqs = hm.generate_density_heatmap(
            (1, 2, 3), (1, 2, 3), (1, np.nan, 1), 1, 1
        )
        npt.assert_array_equal(
            counts, np.array([[0, 0, 1, 1], [0, 0, 1, 1], [1, 1, 0, 0], [1, 1, 0, 0]])
        )



#!/usr/bin/env python
"""This module has tests for the pds.datetime functions."""

# Copyright 2022, United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

import datetime
import unittest

from vipersci.pds import datetime as pdsdt


class TestIsoZ(unittest.TestCase):
    def test_fromisozformat(self):
        dt = datetime.datetime(2022, 10, 1, 13, 20, 0, tzinfo=datetime.timezone.utc)
        self.assertEqual(dt, pdsdt.fromisozformat("2022-10-01T13:20:00Z"))

        self.assertRaises(ValueError, pdsdt.fromisozformat, "2022-10-01T13:20:00")

    def test_isozformat(self):
        dt = datetime.datetime(2022, 10, 1, 13, 20, 0, tzinfo=datetime.timezone.utc)
        self.assertEqual(pdsdt.isozformat(dt), "2022-10-01T13:20:00Z")

        no_tz = dt.replace(tzinfo=None)

        self.assertRaises(ValueError, pdsdt.isozformat, no_tz)



#!/usr/bin/env python
"""This module has tests for the nss_modeler module."""

# Copyright 2022, United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

import argparse
from pathlib import Path
import unittest
from unittest.mock import mock_open, patch

import numpy as np

from vipersci.carto import nss_modeler as nm


class TestParser(unittest.TestCase):
    def test_parser(self):
        parser = nm.arg_parser()
        self.assertIsInstance(parser, argparse.ArgumentParser)


class TestWrite(unittest.TestCase):
    def test_write_tif(
        self,
    ):
        m = mock_open()
        arr = np.array([1, 2])
        path = Path("dummy")
        ending = "end.tif"
        kwds = {"foo": "bar"}
        with patch("vipersci.carto.nss_modeler.rasterio.open", m):
            nm.write_tif(path, ending, arr, kwds)

        m.assert_called_once_with(path.with_name(path.name + ending), "w", **kwds)
        handle = m()
        handle.write.assert_called_once_with(arr, 1)



#!/usr/bin/env python
# coding: utf-8

"""Defines the mapping or join table connecting the image_requests and
ldst tables using the SQLAlchemy ORM."""

# Copyright 2023, United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

from sqlalchemy import Boolean, ForeignKey, Integer, String
from sqlalchemy.orm import mapped_column, relationship

from vipersci.vis.db import Base
from vipersci.vis.db.image_requests import ImageRequest
from vipersci.vis.db.ldst import LDST


class JuncImageRequestLDST(Base):
    """An object to represent the junction of image_requests to LDST hypotheses."""

    __tablename__ = "junc_image_request_ldst"

    image_request_id = mapped_column(
        Integer, ForeignKey(ImageRequest.id), nullable=False, primary_key=True
    )
    ldst_id = mapped_column(
        String, ForeignKey(LDST.id), nullable=False, primary_key=True
    )
    critical = mapped_column(
        Boolean,
        doc="If true, indicates that this ImageRequest is critical for this LDST "
        "hypothesis.",
    )
    evaluation = mapped_column(
        Boolean,
        nullable=True,
        doc="True if a Scientist has indicated that the acquired images for the "
        "indicated ImageRequest supports the LDST Hypothesis.  If false, a Scientist "
        "has determined that the images do not support the LDST Hypothesis.",
    )
    evaluation_notes = mapped_column(
        String,
        nullable=True,
        doc="Any notes about the evaluation of the ImageRequest's images against the "
        "LDST hypothesis.",
    )
    evaluator = mapped_column(
        String,
        nullable=True,
        doc="The name of the individual that reviewed the acquired images against the "
        "LDST hypothesis.",
    )

    image_request = relationship("ImageRequest", back_populates="ldst_associations")
    ldst = relationship("LDST", back_populates="image_request_associations")



#!/usr/bin/env python
"""This module has tests for the image_requests module."""

# Copyright 2022-2023, United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

from datetime import datetime, timezone
import unittest

from vipersci.vis.db import image_requests as ir
from vipersci.vis.db.junc_image_req_ldst import JuncImageRequestLDST  # noqa


class TestEnums(unittest.TestCase):
    def test_Status(self):
        self.assertEqual(ir.Status.WORKING, ir.Status(1))
        self.assertEqual(ir.Status.PLANNED, ir.Status(4))

    def test_CameraType(self):
        self.assertEqual(ir.CameraType.NAVCAM, ir.CameraType(1))
        self.assertEqual(ir.CameraType.AFTCAM, ir.CameraType(2))
        self.assertEqual(ir.CameraType.HAZCAM, ir.CameraType(3))

    def test_CompressionType(self):
        self.assertEqual(ir.CompressionType.LOSSLESS, ir.CompressionType(1))
        self.assertEqual(ir.CompressionType.LOSSY, ir.CompressionType(2))

    def test_ImageMode(self):
        self.assertEqual(ir.ImageMode.LEFT, ir.ImageMode(1))
        self.assertEqual(ir.ImageMode.PANORAMA, ir.ImageMode(4))


class TestImageRequest(unittest.TestCase):
    def setUp(self):
        self.mini = {
            "title": "Mini Request",
            "justification": "Saw a great crater.",
            "status": ir.Status.IMMEDIATE,
            "users": "Rick Elphic",
            "request_time": datetime.now(tz=timezone.utc),
        }

        self.stereo = {
            "title": "Dummy Title",
            "justification": "Greatest Rock of All Time.",
            "status": "WORKING",
            "users": "Ross A. Beyer (VIS), Tony Colaprete",
            "request_time": datetime.now(tz=timezone.utc),
            "target_location": "Rim of the big crater.",
            "rover_location": "A few meters away from the big crater.",
            "rover_orientation": "Any",
            "camera_type": ir.CameraType.NAVCAM,
            "imaging_mode": ir.ImageMode.STEREO,
            "compression": ir.CompressionType.LOSSY,
            "luminaires": "default",
            "exposure_time": "Default",
        }

        self.pano = self.stereo.copy()
        self.pano["imaging_mode"] = ir.ImageMode.PANORAMA
        self.pano.update(
            {
                "caltarget_required": True,
                "aftcam_pair": True,
                "chin_down_navcam_pair": True,
                "slices": 6,
            }
        )

    def test_init(self):
        ir1 = ir.ImageRequest(**self.mini)
        self.assertEqual(ir1.status, ir.Status.IMMEDIATE)

        ir2 = ir.ImageRequest(**self.stereo)
        self.assertEqual(ir2.imaging_mode, ir.ImageMode.STEREO)

        ir3 = ir.ImageRequest(**self.pano)
        self.assertEqual(ir3.slices, 6)

    def test_init_raise(self):
        d = self.stereo.copy()
        d["luminaires"] = "not a luminaire"
        self.assertRaises(ValueError, ir.ImageRequest, **d)

        e = self.stereo.copy()
        e["hazcams"] = "ncl,hfp"
        self.assertRaises(ValueError, ir.ImageRequest, **e)



#!/usr/bin/env python
# coding: utf-8

"""Provides functions to validate various intput data."""

# Copyright 2023, United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

from datetime import datetime, timedelta, timezone

from vipersci.pds import Purpose
from vipersci.pds.datetime import fromisozformat


def validate_datetime_asutc(key, value):
    if isinstance(value, datetime):
        if value.utcoffset() is None:
            raise ValueError(f"{key} must be tz aware.")
        elif value.utcoffset() != timedelta():
            raise ValueError(f"{key} must be tz aware with a UTC offset.")
        dt = value
    elif isinstance(value, str):
        if value.endswith("Z"):
            dt = fromisozformat(value)
        else:
            dt = datetime.fromisoformat(value)
    else:
        raise ValueError(f"{key} must be a datetime or an ISO 8601 formatted string.")

    return dt.astimezone(timezone.utc)


def validate_purpose(value: str):
    s = set(Purpose.__members__.keys())
    if value in s:
        return value

    if value.upper() in s:
        return value.upper()

    raise ValueError(f"purpose must be one of {s}")



#!/usr/bin/env python
"""This module has tests for the pano_records module."""

# Copyright 2023, United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

from datetime import datetime, timedelta, timezone
import unittest

from vipersci.vis.db import pano_records as tpp
from vipersci.vis.db.image_records import ImageRecord  # noqa
from vipersci.vis.db.image_requests import ImageRequest  # noqa
from vipersci.vis.db.junc_image_pano import JuncImagePano  # noqa
from vipersci.vis.db.junc_image_record_tags import JuncImageRecordTag  # noqa
from vipersci.vis.db.junc_image_req_ldst import JuncImageRequestLDST  # noqa


class TestPanoRecord(unittest.TestCase):
    def setUp(self):
        self.startUTC = datetime(2022, 1, 27, 0, 0, 0, tzinfo=timezone.utc)
        self.source_products = ["220127-000000-ncl-d", "220127-000005-ncl-d"]
        self.d = dict(
            file_creation_datetime=datetime.now(timezone.utc),
            file_path="/path/to/dummy",
            lines=2048,
            md5_checksum="dummychecksum",
            mission_phase="Test",
            purpose="Engineering",
            samples=2048,
            source_pids=self.source_products,
            start_time=self.startUTC,
        )
        self.extras = dict(foo="bar")

    def test_init(self):
        p = tpp.PanoRecord(**self.d)
        self.assertEqual("220127-000000-ncl-pan", str(p.product_id))

        d = self.d
        d.update(self.extras)
        ppl = tpp.PanoRecord(**d)
        self.assertEqual("220127-000000-ncl-pan", str(ppl.product_id))

        # Force alternate time
        d = self.d.copy()
        d["start_time"] = self.startUTC + timedelta(hours=1)
        pp2 = tpp.PanoRecord(**d)
        self.assertEqual("220127-010000-ncl-pan", str(pp2.product_id))

        # Remove explicit time signature
        d = self.d.copy()
        del d["start_time"]
        pp3 = tpp.PanoRecord(**d)
        self.assertEqual("220127-000000-ncl-pan", str(pp3.product_id))

        # print(f"{k}: {getattr(rp, k)}")

    def test_init_errors(self):
        d = self.d.copy()
        d["product_id"] = "220127-010000-ncl-b"
        self.assertRaises(ValueError, tpp.PanoRecord, **d)

        d = self.d.copy()
        d["product_id"] = "220127-000000-ncr-b"
        self.assertRaises(ValueError, tpp.PanoRecord, **d)

    def test_product_id(self):
        p = tpp.PanoRecord(**self.d)
        self.assertRaises(NotImplementedError, setattr, p, "product_id", "dummy")

    def test_purpose(self):
        p = tpp.PanoRecord(**self.d)
        self.assertRaises(ValueError, setattr, p, "purpose", "dummy")

    def test_update(self):
        p = tpp.PanoRecord(**self.d)
        k = "foo"
        self.assertTrue(k not in p.labelmeta)

        p.update(self.extras)
        self.assertTrue(k in p.labelmeta)

    # def test_labeldict(self):
    #     din = self.d
    #     din.update(self.extras)
    #     p = tpp.PanoRecord(**din)
    #     d = p.label_dict()
    #     self.assertEqual(d["samples"], p.samples)



#!/usr/bin/env python
"""This module has tests for the vis.create_pano functions."""

# Copyright 2023, vipersci developers.
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

from argparse import ArgumentParser
from datetime import datetime, timezone
from pathlib import Path
import unittest
from unittest.mock import create_autospec, Mock, patch

from geoalchemy2 import load_spatialite
import numpy as np
from sqlalchemy import create_engine
from sqlalchemy.event import listen
from sqlalchemy.orm import Session

from vipersci.vis.db import Base
from vipersci.vis.db.image_records import ImageRecord
from vipersci.vis.db.pano_records import PanoRecord
from vipersci.vis import create_pano as cp


class TestCLI(unittest.TestCase):
    def test_arg_parser(self):
        p = cp.arg_parser()
        self.assertIsInstance(p, ArgumentParser)
        # self.assertRaises(SystemExit, p.parse_args)
        d = vars(
            p.parse_args(
                ["--dburl", "db://foo:username@host/db", "product_id_goes_here"]
            )
        )
        self.assertIn("dburl", d)
        self.assertIn("output_dir", d)
        self.assertIn("inputs", d)

    def test_main(self):
        pa_ret_val = cp.arg_parser().parse_args(
            ["231126-000000-ncl-s.dummy", "231126-000000-ncr-s.dummy"]
        )
        with patch("vipersci.vis.create_pano.arg_parser") as parser, patch(
            "vipersci.vis.create_pano.create"
        ) as m_create:
            parser.return_value.parse_args.return_value = pa_ret_val
            cp.main()
            m_create.assert_called_once()

        pa2_ret_val = cp.arg_parser().parse_args(
            [
                "--dburl",
                "db://foo:username@host/db",
                "231126-000000-ncl-s.dummy",
                "231126-000000-ncr-s.dummy",
            ]
        )
        session_engine_mock = create_autospec(Session)
        session_mock = create_autospec(Session)
        session_engine_mock.__enter__ = Mock(return_value=session_mock)
        with patch("vipersci.vis.create_pano.arg_parser") as parser, patch(
            "vipersci.vis.create_pano.create"
        ) as m_create, patch("vipersci.vis.create_pano.create_engine"), patch(
            "vipersci.vis.create_pano.Session", return_value=session_engine_mock
        ):
            parser.return_value.parse_args.return_value = pa2_ret_val
            cp.main()
            m_create.assert_called_once()
            session_engine_mock.__enter__.assert_called_once()
            session_mock.commit.assert_called_once()


class TestMakePano(unittest.TestCase):
    def setUp(self) -> None:
        self.startUTC = datetime(2022, 1, 27, 0, 0, 0, tzinfo=timezone.utc)
        self.d = {
            "instrument_name": "NavCam Left",
            "lines": 2048,
            "product_id": "230421-200000-ncl-pan",
            "samples": 6000,
        }

    def test_no_image(self):
        pp = cp.make_pano_record(self.d)
        self.assertIsInstance(pp, PanoRecord)

    @patch("vipersci.vis.create_pano.imsave")
    @patch(
        "vipersci.vis.create_pano.tif_info",
        return_value={
            "file_byte_offset": 10,
            "file_creation_datetime": datetime.fromtimestamp(1700921056, timezone.utc),
            "file_data_type": "UnsignedMSB2",
            "file_md5_checksum": "md5",
            "file_path": "dummy.tif",
            "lines": 2048,
            "samples": 2048,
        },
    )
    def test_image(self, mock_tif_info, mock_imsave):
        image = np.array([[5, 5], [5, 5]], dtype=np.uint16)
        pp = cp.make_pano_record(self.d, image, Path("outdir/"))
        self.assertIsInstance(pp, PanoRecord)
        mock_imsave.assert_called_once()
        mock_tif_info.assert_called_once()

        mock_imsave.reset_mock()
        mock_tif_info.reset_mock()

        prp = cp.make_pano_record(self.d, Path("dummy.tif"), Path("outdir/"))
        self.assertIsInstance(prp, PanoRecord)
        mock_imsave.assert_not_called()
        mock_tif_info.assert_called_once()


class TestCreate(unittest.TestCase):
    def test_nodb(self):
        self.assertRaises(ValueError, cp.create, [1, 2, 3])
        self.assertRaises(
            FileNotFoundError,
            cp.create,
            ["231126-000000-ncl-s.dummy", "231126-000000-ncr-s.dummy"],
        )

        path_mock = create_autospec(Path)
        path_mock.exists.return_value = True

        with patch("vipersci.vis.create_pano.Path", return_value=path_mock), patch(
            "vipersci.vis.create_pano.imread"
        ) as m_imread, patch("vipersci.vis.create_pano.np.hstack"), patch(
            "vipersci.vis.create_pano.make_pano_record"
        ) as m_mpr, patch(
            "vipersci.vis.create_pano.write_json"
        ) as m_write_json, patch(
            "vipersci.vis.create_pano.isinstance",
            side_effect=[True, True, False, True, False, True],
        ):
            cp.create(["231126-000000-ncl-s.dummy", "231126-000000-ncr-s.dummy"])

            self.assertEqual(m_imread.call_count, 2)
            self.assertEqual(m_mpr.call_args[0][2], Path.cwd())
            m_write_json.assert_called_once()

    def test_db(self):
        ir1 = ImageRecord(
            adc_gain=0,
            auto_exposure=0,
            cameraId=0,
            capture_id=1,
            exposure_duration=111,
            file_byte_offset=256,
            file_creation_datetime="2023-08-01T00:51:51.148919Z",
            file_data_type="UnsignedLSB2",
            file_md5_checksum="b8f1a035e39c223e2b7e236846102c29",
            file_path="231126-000000-ncl-s",
            imageDepth=2,
            image_id=0,
            immediateDownloadInfo=16,
            instrument_name="NavCam Left",
            instrument_temperature=0,
            lines=2048,
            lobt=1700956800,
            offset=0,
            outputImageType="SLOG_ICER_IMAGE",
            output_image_mask=16,
            padding=0,
            pga_gain=1.0,
            processing_info=26,
            product_id="231126-000000-ncl-s",
            samples=2048,
            software_name="vipersci",
            software_program_name="vipersci.vis.create_image",
            software_version="0.6.0-dev",
            start_time="2023-11-26T00:00:00Z",
            stereo=1,
            stop_time="2023-11-26T00:00:00.000111Z",
            voltage_ramp=0,
            yamcs_generation_time="2023-11-26T00:00:00Z",
            yamcs_name="/ViperGround/Images/ImageData/Navcam_left_slog",
        )
        ir2 = ImageRecord(
            adc_gain=0,
            auto_exposure=0,
            cameraId=1,
            capture_id=2,
            exposure_duration=111,
            file_byte_offset=256,
            file_creation_datetime="2023-08-01T00:51:51.148919Z",
            file_data_type="UnsignedLSB2",
            file_md5_checksum="b8f1a035e39c223e2b7e236846102c29",
            file_path="231126-000000-ncr-s",
            imageDepth=2,
            image_id=0,
            immediateDownloadInfo=16,
            instrument_name="NavCam Right",
            instrument_temperature=0,
            lines=2048,
            lobt=1700956800,
            offset=0,
            outputImageType="SLOG_ICER_IMAGE",
            output_image_mask=16,
            padding=0,
            pga_gain=1.0,
            processing_info=26,
            product_id="231126-000000-ncr-s",
            samples=2048,
            software_name="vipersci",
            software_program_name="vipersci.vis.create_image",
            software_version="0.6.0-dev",
            start_time="2023-11-26T00:00:00Z",
            stereo=1,
            stop_time="2023-11-26T00:00:00.000111Z",
            voltage_ramp=0,
            yamcs_generation_time="2023-11-26T00:00:00Z",
            yamcs_name="/ViperGround/Images/ImageData/Navcam_left_slog",
        )
        engine = create_engine("sqlite:///:memory:")
        listen(engine, "connect", load_spatialite)
        session = Session(engine)
        Base.metadata.create_all(engine)
        session.add(ir1)
        session.add(ir2)
        session.commit()

        path_mock = create_autospec(Path)
        path_mock.exists.return_value = True

        with patch("vipersci.vis.create_pano.Path", return_value=path_mock), patch(
            "vipersci.vis.create_pano.imread"
        ) as m_imread, patch("vipersci.vis.create_pano.np.hstack"), patch(
            "vipersci.vis.create_pano.make_pano_record"
        ) as m_mpr, patch(
            "vipersci.vis.create_pano.isinstance",
            side_effect=[True, True, True, True, True, True],
        ):
            session.add_all = Mock()
            cp.create(
                ["231126-000000-ncl-s", "231126-000000-ncr-s"],
                session=session,
                json=False,
            )

            self.assertEqual(m_imread.call_count, 2)
            self.assertEqual(m_mpr.call_args[0][2], Path.cwd())



#!/usr/bin/env python
"""This module contains classes for VIPER Product IDs."""

# Copyright 2022-2024, United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

import datetime
from itertools import groupby
import re
from typing import Iterable

instruments = dict()
vis_instruments = dict(
    ncl="NavCam Left",
    ncr="NavCam Right",
    hfp="HazCam Forward Port",
    hfs="HazCam Forward Starboard",
    hap="HazCam Aft Port",
    has="HazCam Aft Starboard",
    acl="AftCam Left",
    acr="AftCam Right",
    pan="Panorama",  # Not an instrument, but a valid combined product type.
)
vis_instrument_aliases = {
    "navcam left": "ncl",
    "navcam right": "ncr",
    "aftcam left": "acl",
    "aftcam right": "acr",
    "hazcam forward port": "hfp",
    "hazcam forward starboard": "hfs",
    "hazcam aft port": "hap",
    "hazcam aft starboard": "has",
    "hazcam front left": "hfp",
    "hazcam front right": "hfs",
    "hazcam back left": "hap",
    "hazcam back right": "has",
    "hazcam_1": "hfp",
    "hazcam_2": "hap",
    "hazcam_3": "hfs",
    "hazcam_4": "has",
}
vis_instrument_numbers = {
    0: "ncl",
    1: "ncr",
    2: "acl",
    3: "acr",
    4: "hfp",
    5: "hfs",
    6: "hap",
    7: "has",
}
instruments.update(vis_instruments)
vis_compression = dict(
    a=1,  # 1:1 Lossless compression
    b=5,  # 5:1 compression
    c=16,  # 16:1 compression
    d=24,  # 24:1 compression
    s="SLoG",  # SLoG compression
    z=None,  # Uncompressed
)

nirvss_instruments = dict(
    aim="Ames Imaging Module",
)
instruments.update(nirvss_instruments)

# Create some compiled regex Patterns to use in this module.
date_re = re.compile(r"(2\d)(1[0-2]|0[1-9])(3[01]|[12][0-9]|0[1-9])")  # YYMMDD
time_re = re.compile(r"(2[0-3]|[01]\d)([0-5]\d)([0-5]\d)(\d{3})?")  # hhmmssfff
inst_re = re.compile("|".join(instruments.keys()))
vis_inst_re = re.compile("|".join(vis_instruments.keys()))

pid_re = re.compile(
    rf"(?P<date>{date_re.pattern})-"
    rf"(?P<time>{time_re.pattern})-"
    rf"(?P<instrument>{inst_re.pattern})"
)

vis_comp_re = re.compile("|".join(vis_compression.keys()))
vis_pid_re = re.compile(pid_re.pattern + rf"-(?P<compression>{vis_comp_re.pattern})")
vis_pan_re = re.compile(
    f"(?P<date>{date_re.pattern})-"
    rf"(?P<time>{time_re.pattern})-"
    rf"((?P<instrument>{vis_inst_re.pattern})-)?"
    rf"pan"
)


def get_key(value, dictionary):
    for k, v in dictionary.items():
        if v == value:
            return k
    else:
        raise KeyError(f"No value, {value}, was found in the dictionary.")


class VIPERID:
    """A Class for VIPER Product IDs.

    :ivar date: a six digit string denoting YYMMDD (or strftime %y%m%d) where
        the two digit year can be prefixed with "20" to get the four-digit year.
    :ivar time: a six or nine digit string denoting hhmmss (or strftime
        %H%M%S%f) or hhmmssuuu, similar to the first, but where the trailing
        three digits are miliseconds.
    :ivar instrument: A three character sequence denoting the instrument.
    """

    def __init__(self, *args):
        if len(args) == 1:
            match = pid_re.search(str(args).lower())
            if match:
                parsed = match.groupdict()
                self.date = parsed["date"]
                self.time = parsed["time"]
                self.instrument = parsed["instrument"]
            else:
                raise ValueError(f"{args} did not match regex: {pid_re.pattern}")
        else:
            if len(args) == 2:
                if isinstance(args[0], datetime.datetime):
                    date = args[0].date()
                    time = args[0].time()
                else:
                    raise ValueError(
                        "For two arguments, the first must be a datetime" "object."
                    )
                instrument = args[1]

            elif len(args) == 3:
                if isinstance(args[0], (datetime.date, str)):
                    date = args[0]
                else:
                    raise ValueError(
                        "For three arguments, the first must be a date or "
                        "string object."
                    )

                if isinstance(args[1], (datetime.time, str)):
                    time = args[1]
                else:
                    raise ValueError(
                        "For three arguments, the second must be a time or "
                        "string object."
                    )

                instrument = args[2]

            else:
                raise IndexError("accepts 1 to 3 arguments")

            self.date = self.format_date(date)
            self.time = self.format_time(time)

            match = inst_re.search(instrument)
            if match:
                self.instrument = match[0]
            elif instrument in instruments.values():
                self.instrument = get_key(instrument, instruments)
            else:
                raise ValueError(f"{instrument} did not match regex: {inst_re.pattern}")
        return

    def __str__(self):
        return "-".join((self.date, self.time, self.instrument))

    def __repr__(self):
        return f"{self.__class__.__name__}('{self.__str__()}')"

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return (
                self.date == other.date
                and self.time == other.time
                and self.instrument == other.instrument
            )
        return False

    def __hash__(self):
        return hash((self.date, self.time, self.instrument))

    def __lt__(self, other):
        if isinstance(other, self.__class__):
            return (
                self.date,
                self.time,
                self.instrument,
            ) < (
                other.date,
                other.time,
                other.instrument,
            )
        else:
            return NotImplemented

    @staticmethod
    def format_date(date) -> str:
        if isinstance(date, datetime.date):
            if datetime.date(2000, 1, 1) <= date < datetime.date(2100, 1, 1):
                datestr = date.strftime("%y%m%d")
            else:
                raise ValueError("Date must be between the year 2000 and 2100.")
        else:
            match = date_re.search(date)
            if match:
                datestr = match[0]
            else:
                raise ValueError(f"{date} did not match regex: {date_re.pattern}")

        return datestr

    @staticmethod
    def format_time(time) -> str:
        if isinstance(time, datetime.time):
            if time.microsecond == 0:
                timestr = time.strftime("%H%M%S")
            elif time.microsecond >= 1000:
                timestr = time.strftime("%H%M%S%f")[:9]
            else:
                raise ValueError(
                    "The provided time has more precision "
                    "than a milisecond, which is not allowed for "
                    "VIPER IDs."
                )
        else:
            match = time_re.search(time)
            if match:
                timestr = match[0]
            else:
                raise ValueError(f"{time} did not match regex: {time_re.pattern}")

        return timestr

    def datetime(self):
        fmt = "%y%m%d-%H%M%S"
        time_string = f"{self.date}-{self.time}"
        if len(self.time) == 9:
            fmt += "%f"
            time_string += "000"
        dt = datetime.datetime.strptime(time_string, fmt)
        return dt.replace(tzinfo=datetime.timezone.utc)


class VISID(VIPERID):
    """A Class for VIPER VIS Product IDs.

    :ivar date: a six digit string denoting YYMMDD (or strftime %y%m%d) where
        the two digit year can be prefixed with "20" to get the four-digit year.
    :ivar time: a six or nine digit string denoting hhmmss (or strftime
        %H%M%S%f) or hhmmssuuu, similar to the first, but where the trailing
        three digits are miliseconds.
    :ivar instrument: A three character sequence denoting the instrument.
    """

    def __init__(self, *args):
        if len(args) == 1:
            if isinstance(args[0], dict):
                if "start_time" in args[0] or "lobt" in args[0]:
                    if args[0].keys() >= {"start_time", "lobt"}:
                        if (
                            datetime.datetime.fromtimestamp(
                                args[0]["lobt"], tz=datetime.timezone.utc
                            )
                            != args[0]["start_time"]
                        ):
                            raise ValueError(
                                f"The start_time {args[0]['start_time']} does not "
                                f"equal the lobt {args[0]['lobt']}"
                            )
                    if "lobt" in args[0]:
                        dt = datetime.datetime.fromtimestamp(
                            args[0]["lobt"], tz=datetime.timezone.utc
                        )
                    else:
                        dt = args[0]["start_time"]

                    date = dt.date()
                    time = dt.time()
                else:
                    raise ValueError(
                        "The dictionary had neither 'start_time' nor 'lobt' " "keys."
                    )
                instrument = args[0]["instrument_name"]
                compression = args[0]["onboard_compression_ratio"]
            else:
                match = vis_pid_re.search(str(args).lower())
                if match:
                    parsed = match.groupdict()
                    date = parsed["date"]
                    time = parsed["time"]
                    instrument = parsed["instrument"]
                    compression = parsed["compression"]
                else:
                    raise ValueError(
                        f"{args} did not match regex: {vis_pid_re.pattern}"
                    )
        elif len(args) == 4:
            (date, time, instrument, compression) = args
        else:
            raise IndexError("accepts 1 or 4 arguments")

        instrument = self.instrument_name(instrument)

        super().__init__(date, time, instrument)
        self.compression = self.compression_letter(compression)

    def __str__(self):
        return "-".join((super().__str__(), self.compression))

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return super().__eq__(other) and self.compression == other.compression
        return False

    def __hash__(self):
        return hash((super().__hash__(), self.compression))

    def __lt__(self, other):
        if isinstance(other, self.__class__):
            if super().__eq__(other):
                if "z" in (self.compression, other.compression):
                    if other.compression == "z" and self.compression != "z":
                        return False
                    else:
                        return True
                else:
                    return self.compression < other.compression
            else:
                return super().__lt__(other)
        else:
            return NotImplemented

    @staticmethod
    def instrument_name(name):
        """Returns fullname of VIS instrument based on *name*."""
        if isinstance(name, int):
            return vis_instruments[vis_instrument_numbers[name]]
        if name.casefold() in vis_instruments:
            return vis_instruments[name.casefold()]
        elif name.casefold() in vis_instrument_aliases:
            return vis_instruments[vis_instrument_aliases[name.casefold()]]
        else:
            for k in vis_instrument_aliases.keys():
                if k in name.casefold():
                    return vis_instruments[vis_instrument_aliases[k]]
            else:
                raise ValueError(
                    f"No VIS instrument name based on {name} could be found."
                )

    def compression_class(self):
        """Returns text value for the PDS onboard_compression_class."""
        if self.compression == "a" or self.compression == "s":
            return "Lossless"
        else:
            return "Lossy"

    @staticmethod
    def compression_letter(compression):
        """Returns the letter code from the pid.vis_compression dictionary that matches
        the value provided via *compression*.
        """
        if compression in vis_compression:
            return compression
        elif compression in vis_compression.values():
            return get_key(compression, vis_compression)
        elif isinstance(compression, (int, float)):
            compression_ratios = []
            for v in vis_compression.values():
                if isinstance(v, (int, float)):
                    compression_ratios.append(v)

            if len(compression_ratios) == 0:
                raise ValueError(
                    "There are no numeric values in vis_compression "
                    f"({vis_compression})."
                )

            for r in sorted(compression_ratios, reverse=True):
                if compression >= r:
                    return get_key(r, vis_compression)
            else:
                raise ValueError(
                    f"The numeric value of {compression} is not greater than one "
                    f"of {compression_ratios}"
                )
        else:
            raise ValueError(
                f"Could not determine one of {vis_compression.keys()} from "
                f"compression ({compression})."
            )

    @staticmethod
    def best_compression(identifiers: Iterable):
        """
        Returns a list of elements in *identifiers* eliminating identifiers which
        can be made into VISIDs which come from the same observation, but have poorer
        compression.

        For example:
        >>> best = best_compression(["241127-010203-ncl-a", "241127-010203-ncl-b"])
        >>> print(best)
        ["241127-010203-ncl-a",]
        """
        d = dict()
        for i in identifiers:
            d[VISID(i)] = i

        best_ids = list()
        vids = sorted(d.keys())
        for k, g in groupby(vids, key=VIPERID):
            best_ids.append(sorted(g)[0])

        return [d[x] for x in best_ids]


class PanoID(VIPERID):
    """A Class for VIPER VIS Panorama Product IDs.  The date/time combination
       should be equal to the earliest source product date/time combination.

    :ivar date: a six digit string denoting YYMMDD (or strftime %y%m%d) where
        the two digit year can be prefixed with "20" to get the four-digit year.
    :ivar time: a six or nine digit string denoting hhmmss (or strftime
        %H%M%S%f) or hhmmssuuu, similar to the first, but where the trailing
        three digits are miliseconds.
    :ivar instrument: A three character sequence denoting the instrument (if all
        source data came from the same instrument), can be None (indicating multiple
        instruments contributed).
    """

    def __init__(self, *args):
        if len(args) == 1:
            match = vis_pan_re.search(str(args).lower())
            if not match:
                raise ValueError(f"{args} did not match regex: {vis_pan_re.pattern}")
        elif 2 <= len(args) <= 3:
            instrument = args[-1]
            VISID.instrument_name(instrument)

        super().__init__(*args)
        if self.instrument == "pan":
            self.instrument = None

    def __str__(self):
        if self.instrument in ("pan", None):
            return "-".join((self.date, self.time, "pan"))
        else:
            return "-".join((self.date, self.time, self.instrument, "pan"))



#!/usr/bin/env python
"""This module has tests for the pds.pid functions."""

# Copyright 2022, United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

# import contextlib
# import shutil
import datetime
import unittest
from unittest.mock import patch

# from pathlib import Path

from vipersci.pds import pid


class TestGetKey(unittest.TestCase):
    def test_get_key(self):
        d = {"a": "alpha", "b": "beta", "g": "gamma"}
        self.assertEqual(pid.get_key("beta", d), "b")
        self.assertRaises(KeyError, pid.get_key, "omega", d)


class TestVIPERID(unittest.TestCase):
    def test_init_tuple(self):
        tuples = (
            (
                ("220117", "010101", "ncl"),
                (datetime.date(2022, 1, 17), datetime.time(1, 1, 1), "ncl"),
            ),
            (
                ("220117", "010101", "ncl"),
                (datetime.date(2022, 1, 17), datetime.time(1, 1, 1), "NavCam Left"),
            ),
            (
                ("220117", "010101001", "aim"),
                (datetime.datetime(2022, 1, 17, 1, 1, 1, 1000), "aim"),
            ),
            (
                ("231225", "182000", "acr"),
                (datetime.date(2023, 12, 25), datetime.time(18, 20, 0), "acr"),
            ),
            (
                ("240330", "121212", "hap"),
                (datetime.datetime(2024, 3, 30, 12, 12, 12), "hap"),
            ),
        )
        for truth, t in tuples:
            with self.subTest():
                p = pid.VIPERID(*t)
                self.assertTupleEqual(truth, (p.date, p.time, p.instrument))

    def test_init_string(self):
        string_tuples = (
            ("220117-010101-ncl", "220117-010101-ncl"),
            ("220117-010101-ncl", "220117-010101-NCL"),
            ("220117-010101-ncl", "20220117-010101-ncl"),
            ("231225-182000-acr", "231225-182000-acr ignored"),
        )
        for s in string_tuples:
            with self.subTest():
                p = pid.VIPERID(s[1])
                self.assertEqual(s[0], str(p))

    def test_init_bad_tuples(self):
        tuples = (
            ("not a datetime", "ncl"),
            ("not a date", datetime.time(1, 1, 1), "ncl"),
            (datetime.date(2024, 1, 1), "not a time", "ncl"),
            (datetime.datetime(2022, 1, 17, 1, 1, 1), "not an instrument"),
            (999, datetime.time(1, 1, 1), "ncl"),
            (datetime.date(2024, 1, 1), 999, "ncl"),
        )
        for t in tuples:
            with self.subTest(t):
                self.assertRaises(ValueError, pid.VIPERID, *t)

    def test_init_bad_strings(self):
        strings = (
            "220117-010101-aaa",
            "220117-10101-ncl",
            "foobar",
        )
        for s in strings:
            with self.subTest(s):
                self.assertRaises(ValueError, pid.VIPERID, s)

    def test_init_too_many_args(self):
        self.assertRaises(
            IndexError,
            pid.VIPERID,
            datetime.date(2024, 1, 1),
            datetime.time(1, 1, 1),
            "aim",
            "more than three arguments",
        )
        self.assertRaises(
            IndexError,
            pid.VIPERID,
        )

    def test_init_zero_args(self):
        self.assertRaises(IndexError, pid.VIPERID)

    def test_repr(self):
        s = "This is a VIPER ID: 220117-010101-ncl"
        p = pid.VIPERID(s)
        self.assertEqual("VIPERID('220117-010101-ncl')", repr(p))

    def test_str_(self):
        test = "This is a VIPER Id: 220117-010101-aim"
        truth = "220117-010101-aim"
        self.assertEqual(truth, pid.VIPERID(test).__str__())

    def test_eq(self):
        p1 = pid.VIPERID("231120-010101-acl")
        self.assertFalse(p1 == "foo")

    def test_lt(self):
        p1 = pid.VIPERID("231120-010101-acl")
        p2 = pid.VIPERID("231121-010101-acl")
        p3 = pid.VIPERID("231121-010102-acl")
        p4 = pid.VIPERID("231121-010102-ncl")
        self.assertTrue(p1 < p2)
        self.assertTrue(p2 < p3)
        self.assertTrue(p3 < p4)
        self.assertEqual(NotImplemented, p1.__lt__("foo"))

    def test_format_time(self):
        self.assertRaises(ValueError, pid.VISID.format_time, datetime.time(1, 1, 1, 1))

    def test_datetime(self):
        p = pid.VIPERID("231121-010101-acl")
        self.assertEqual(
            p.datetime(),
            datetime.datetime(2023, 11, 21, 1, 1, 1, tzinfo=datetime.timezone.utc),
        )

        p2 = pid.VIPERID("240219-010101001-aim")
        self.assertEqual(
            p2.datetime(),
            datetime.datetime(2024, 2, 19, 1, 1, 1, 1000, tzinfo=datetime.timezone.utc),
        )


class TestVISID(unittest.TestCase):
    def test_init_tuple(self):
        tuples = (
            (
                ("220117", "010101", "ncl", "a"),
                (datetime.date(2022, 1, 17), datetime.time(1, 1, 1), "ncl", "a"),
            ),
            (
                ("220117", "010101", "ncl", "a"),
                (datetime.date(2022, 1, 17), datetime.time(1, 1, 1), 0, "a"),
            ),
            (
                ("220117", "010101", "ncl", "b"),
                (datetime.date(2022, 1, 17), datetime.time(1, 1, 1), "ncl", 5),
            ),
            (
                ("240330", "121212", "hap", "d"),
                (datetime.date(2024, 3, 30), datetime.time(12, 12, 12), "hap", "d"),
            ),
        )
        for truth, t in tuples:
            with self.subTest(truth=truth, test=t):
                vid = pid.VISID(*t)
                self.assertTupleEqual(
                    truth,
                    (
                        vid.date,
                        vid.time,
                        vid.instrument,
                        vid.compression,
                    ),
                )

    def test_init_string(self):
        string_tuples = (
            ("220117-010101-ncl-a", "220117-010101-ncl-a"),
            ("220117-010101-ncl-b", "220117-010101-NCL-b"),
            ("231225-182000-acr-d", "231225-182000-acr-d ignored"),
        )
        for s in string_tuples:
            vid = pid.VISID(s[1])
            with self.subTest(truth=s[0], ccid=vid):
                self.assertEqual(s[0], str(vid))

    def test_init_dict(self):
        d = dict(
            lobt=1643241600, instrument_name="NavCam Left", onboard_compression_ratio=5
        )
        vid = pid.VISID(d)
        self.assertEqual("220127-000000-ncl-b", str(vid))

        vid2 = pid.VISID(
            dict(
                start_time=datetime.datetime(
                    2024, 2, 19, 1, 1, 1, tzinfo=datetime.timezone.utc
                ),
                instrument_name="NavCam Left",
                onboard_compression_ratio=5,
            )
        )
        self.assertEqual("240219-010101-ncl-b", str(vid2))

        self.assertRaises(ValueError, pid.VISID, dict(a="NavCam Left", b=5))

    def test_init_bad_tuples(self):
        tuples = (
            (datetime.date(2024, 1, 1), datetime.time(1, 1, 1), "ncl", "w"),
            (datetime.date(2024, 1, 1), datetime.time(1, 1, 1), "ncl", 0.5),
        )
        for t in tuples:
            with self.subTest(test=t):
                self.assertRaises(ValueError, pid.VISID, *t)

        with patch(
            "vipersci.pds.pid.vis_compression", return_value=dict(s="SLoG", z=None)
        ):
            self.assertRaises(
                ValueError,
                pid.VISID,
                dict(
                    lobt=1643241600,
                    instrument_name="NavCam Left",
                    onboard_compression_ratio=5,
                ),
            )

    def test_init_bad_strings(self):
        strings = (
            "220117-010101-ncl",
            "220117-010101-aaa-a",
            "220117-010101-ncl-q",
            "foobar",
        )
        for s in strings:
            with self.subTest(test=s):
                self.assertRaises(ValueError, pid.VISID, s)

    def test_init_bad_dict(self):
        dicts = (
            dict(lobt=0, instrument_name="NavCam Left", onboard_compression_ratio=5),
            dict(
                lobt=1643241600,
                start_time=datetime.datetime(
                    2023, 1, 27, 0, 0, 0, tzinfo=datetime.timezone.utc
                ),
                instrument_name="NavCam Left",
                onboard_compression_ratio=5,
            ),
        )
        for d in dicts:
            with self.subTest(test=d):
                self.assertRaises(ValueError, pid.VISID, d)

    def test_init_wrong_arg_count(self):
        # (datetime.date(2024, 1, 1), datetime.time(1, 1, 1), "ncl")
        self.assertRaises(
            IndexError,
            pid.VISID,
            "1",
            "2",
            "3",
            "4",
            "5",
        )
        self.assertRaises(IndexError, pid.VISID, "1", "2")
        self.assertRaises(
            IndexError,
            pid.VISID,
        )

    def test_repr(self):
        s = "This is a VISID: 220117-010101-ncl-a"
        cid = pid.VISID(s)
        self.assertEqual("VISID('220117-010101-ncl-a')", repr(cid))

    def test_eq(self):
        vid0 = pid.VISID("220117-010101-ncl-z")
        self.assertFalse(vid0 == "foo")

    def test_lt(self):
        vid0 = pid.VISID("220117-010101-ncl-z")
        vid1 = pid.VISID("220117-010101-ncl-a")
        vid2 = pid.VISID("220117-010101-ncl-b")
        vid3 = pid.VISID("220117-010101-ncl-c")
        vid4 = pid.VISID("220117-010101-ncl-d")
        self.assertTrue(vid0 < vid1)
        self.assertTrue(vid1 < vid2)
        self.assertTrue(vid2 < vid3)
        self.assertTrue(vid3 < vid4)

        self.assertEqual(NotImplemented, vid0.__lt__("foo"))

        vids = [vid3, vid4, vid1, vid2, vid0]
        self.assertEqual(sorted(vids), [vid0, vid1, vid2, vid3, vid4])

    def test_instrument_name(self):
        self.assertEqual("NavCam Left", pid.VISID.instrument_name("navcam left orig"))

    def test_best_compression(self):
        truth = ["241127-010203-ncl-a", "241127-010204-ncr-z"]

        test = [
            "241127-010203-ncl-a",
            "241127-010203-ncl-d",
            "241127-010203-ncl-s",
            "241127-010204-ncr-z",
            "241127-010204-ncr-s",
        ]
        result = pid.VISID.best_compression(test)
        self.assertEqual(result, truth)

    def test_compression_class(self):
        vid = pid.VISID("240219-010101-ncl-a")
        self.assertEqual(vid.compression_class(), "Lossless")


class TestPanoID(unittest.TestCase):
    def test_init_tuple(self):
        tuples = (
            (
                ("220117", "010101", "ncl"),
                (datetime.date(2022, 1, 17), datetime.time(1, 1, 1), "ncl"),
            ),
            (
                ("220117", "010101", "ncl"),
                (datetime.date(2022, 1, 17), datetime.time(1, 1, 1), "NavCam Left"),
            ),
            (
                ("231225", "182000", "acr"),
                (datetime.date(2023, 12, 25), datetime.time(18, 20, 0), "acr"),
            ),
            (
                ("240330", "121212", "hap"),
                (datetime.datetime(2024, 3, 30, 12, 12, 12), "hap"),
            ),
        )
        for truth, t in tuples:
            with self.subTest():
                p = pid.PanoID(*t)
                self.assertTupleEqual(truth, (p.date, p.time, p.instrument))

    def test_init_string(self):
        string_tuples = (
            ("220117-010101-ncl-pan", "220117-010101-ncl-pano"),
            ("220117-010101-ncl-pan", "220117-010101-NCL-PAN"),
            ("220117-010101-ncl-pan", "20220117-010101-ncl-pano"),
            ("231225-182000-acr-pan", "231225-182000-acr-pan ignored"),
            ("231225-182000-pan", "231225-182000-pan multiple instruments"),
        )
        for s in string_tuples:
            with self.subTest():
                p = pid.PanoID(s[1])
                self.assertEqual(s[0], str(p))

    def test_init_bad_tuples(self):
        tuples = (
            ("not a datetime", "ncl"),
            ("not a date", datetime.time(1, 1, 1), "ncl"),
            ("220117", "010101001", "aim"),
            (datetime.date(2024, 1, 1), "not a time", "ncl"),
            (datetime.datetime(2022, 1, 17, 1, 1, 1), "not an instrument"),
        )
        for t in tuples:
            with self.subTest(t):
                self.assertRaises(ValueError, pid.PanoID, *t)

    def test_init_bad_strings(self):
        strings = (
            "220117-010101-ncl",
            "220117-10101-ncl",
            "foobar",
        )
        for s in strings:
            with self.subTest(s):
                self.assertRaises(ValueError, pid.PanoID, s)

    def test_repr(self):
        s = "This is a Pano ID: 220117-010101-ncl-pan"
        p = pid.PanoID(s)
        self.assertEqual("PanoID('220117-010101-ncl-pan')", repr(p))

    def test_str_(self):
        test = "This is a Pano Id: 220117-010101-ncr-pan"
        truth = "220117-010101-ncr-pan"
        self.assertEqual(truth, pid.PanoID(test).__str__())



#!/usr/bin/env python
"""This module has tests for the vis.pds.create_raw functions."""

# Copyright 2022-2023, vipersci developers.
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

from argparse import ArgumentParser
from pathlib import Path
import unittest
from unittest.mock import patch

from geoalchemy2 import load_spatialite
from sqlalchemy import create_engine
from sqlalchemy.event import listen
from sqlalchemy.orm import Session

from vipersci.vis.db import Base
from vipersci.vis.db.image_records import ImageRecord
from vipersci.vis.db.junc_image_pano import JuncImagePano
from vipersci.vis.db.pano_records import PanoRecord
from vipersci.vis.pds import create_pano_product as cpp

from datetime_sqlite import isozformat


class TestParser(unittest.TestCase):
    def test_arg_parser(self):
        p = cpp.arg_parser()
        self.assertIsInstance(p, ArgumentParser)
        self.assertRaises(SystemExit, p.parse_args)
        d = vars(
            p.parse_args(
                ["--dburl", "db://foo:username@host/db", "product_id_goes_here"]
            )
        )
        self.assertIn("dburl", d)
        self.assertIn("template", d)
        self.assertIn("tiff", d)
        self.assertIn("output_dir", d)
        self.assertIn("input", d)


class TestDatabase(unittest.TestCase):
    def setUp(self) -> None:
        self.pr = PanoRecord(
            file_creation_datetime="2023-11-27T22:19:47.216057Z",
            file_md5_checksum="9e9e7de3e8943ca2b355e5a5ffd47858",
            file_path="231109-170000-ncl-pan.tif",
            lines=2048,
            product_id="231109-170000-ncl-pan",
            rover_pan_max=180.0,
            rover_pan_min=-180.0,
            rover_tilt_max=15,
            rover_tilt_min=-50,
            samples=12288,
            software_name="vipersci",
            software_program_name="vipersci.vis.create_pano",
            software_version="0.7.0-dev",
            source_pids=[
                "231109-170000-ncl-c",
                "231109-170100-ncl-c",
                "231109-170200-ncl-c",
            ],
            start_time="2023-11-09T17:00:00Z",
            stop_time="2023-11-09T17:05:00.000511Z",
        )
        ir1 = ImageRecord(
            adc_gain=0,
            auto_exposure=0,
            byteQuota=341,
            cameraId=0,
            capture_id=1,
            exposure_duration=511,
            file_byte_offset=256,
            file_creation_datetime="2023-11-27T22:18:11.879458Z",
            file_data_type="UnsignedLSB2",
            file_md5_checksum="cd30229a7803ec35fbf21a3da254ad10",
            file_path="231109-170000-ncl-d.tif",
            imageDepth=2,
            image_id=0,
            immediateDownloadInfo=24,
            instrument_name="NavCam Left",
            instrument_temperature=0,
            lines=2048,
            lobt=1699549200,
            offset=0,
            output_image_mask=8,
            padding=0,
            pga_gain=1.0,
            processing_info=10,
            product_id="231109-170000-ncl-d",
            samples=2048,
            software_name="vipersci",
            software_program_name="vipersci.vis.create_image",
            software_version="0.7.0-dev",
            start_time="2023-11-09T17:00:00Z",
            stereo=1,
            stop_time="2023-11-09T17:02:00.000511Z",
            voltage_ramp=0,
            yamcs_generation_time="2023-11-09T17:00:00Z",
            yamcs_name="/ViperGround/Images/ImageData/Navcam_left_icer",
            yamcs_reception_time="2023-11-09T17:05:00Z",
        )
        ir2 = ImageRecord(
            adc_gain=0,
            auto_exposure=0,
            byteQuota=341,
            cameraId=0,
            capture_id=1,
            exposure_duration=511,
            file_byte_offset=256,
            file_creation_datetime="2023-11-27T22:18:12.948617Z",
            file_data_type="UnsignedLSB2",
            file_md5_checksum="781d510b1f7f7a4048f7a9eea596b9e6",
            file_path="231109-170100-ncl-d.tif",
            imageDepth=2,
            image_id=0,
            immediateDownloadInfo=24,
            instrument_name="NavCam Left",
            instrument_temperature=0,
            lines=2048,
            lobt=1699549260,
            offset=0,
            output_image_mask=8,
            padding=0,
            pga_gain=1.0,
            processing_info=10,
            product_id="231109-170100-ncl-d",
            samples=2048,
            software_name="vipersci",
            software_program_name="vipersci.vis.create_image",
            software_version="0.7.0-dev",
            start_time="2023-11-09T17:01:00Z",
            stereo=1,
            stop_time="2023-11-09T17:01:00.000511Z",
            voltage_ramp=0,
            yamcs_generation_time="2023-11-09T17:01:00Z",
            yamcs_name="/ViperGround/Images/ImageData/Navcam_left_icer",
            yamcs_reception_time="2023-11-09T17:06:00Z",
        )
        ir3 = ImageRecord(
            adc_gain=0,
            auto_exposure=0,
            byteQuota=341,
            cameraId=0,
            capture_id=1,
            exposure_duration=511,
            file_byte_offset=256,
            file_creation_datetime="2023-11-27T22:18:14.068725Z",
            file_data_type="UnsignedLSB2",
            file_md5_checksum="ee7a86d7ed7a436d23dd6b87ae6c3058",
            file_path="231109-170200-ncl-d.tif",
            imageDepth=2,
            image_id=0,
            immediateDownloadInfo=24,
            instrument_name="NavCam Left",
            instrument_temperature=0,
            lines=2048,
            lobt=1699549320,
            offset=0,
            output_image_mask=8,
            padding=0,
            pga_gain=1.0,
            processing_info=10,
            product_id="231109-170200-ncl-d",
            samples=2048,
            software_name="vipersci",
            software_program_name="vipersci.vis.create_image",
            software_version="0.7.0-dev",
            start_time="2023-11-09T17:02:00Z",
            stereo=1,
            stop_time="2023-11-09T17:02:00.000511Z",
            voltage_ramp=0,
            yamcs_generation_time="2023-11-09T17:02:00Z",
            yamcs_name="/ViperGround/Images/ImageData/Navcam_left_icer",
            yamcs_reception_time="2023-11-09T17:07:00Z",
        )
        self.engine = create_engine("sqlite:///:memory:")
        listen(self.engine, "connect", load_spatialite)
        self.session = Session(self.engine)
        Base.metadata.create_all(self.engine)
        to_add = [
            self.pr,
        ]
        for ir in (ir1, ir2, ir3):
            self.session.add(ir)
            a = JuncImagePano()
            a.image_record = ir
            a.pano_record = self.pr
            to_add.append(a)

        self.session.add_all(to_add)
        self.session.commit()

    def tearDown(self):
        Base.metadata.drop_all(self.engine)

    def test_label_dict(self):
        d = cpp.label_dict(self.pr)
        self.assertEqual(
            d["lid"], f"urn:nasa:pds:viper_vis:data_derived:{self.pr.product_id}"
        )
        self.assertEqual(d["source_product_type"], "data_to_raw_source_product")
        self.assertEqual(d["instruments"][0]["name"], "NavCam Left")

    @patch("vipersci.vis.pds.create_pano_product.create_engine")
    @patch("vipersci.vis.pds.create_pano_product.write_xml")
    def test_main(self, m_write_xml, m_create_engine):
        with patch(
            "vipersci.vis.pds.create_pano_product.Session", return_value=self.session
        ):
            pa_ret_val = cpp.arg_parser().parse_args(
                [
                    "--dburl",
                    "db://foo:username@host/db",
                    self.pr.product_id,
                ]
            )
            with patch("vipersci.vis.pds.create_pano_product.arg_parser") as parser:
                parser.return_value.parse_args.return_value = pa_ret_val
                with patch("vipersci.vis.db.image_records.isozformat", new=isozformat):
                    cpp.main()
                    m_write_xml.assert_called_once()
                    (metadata, template, outdir) = m_write_xml.call_args[0]
                    self.assertEqual(metadata["product_id"], self.pr.product_id)
                    self.assertEqual(outdir, Path.cwd())
                    self.assertEqual(template, "pano-template.xml")



#!/usr/bin/env python
"""This module has tests for the anaglyph module."""

# Copyright 2023, United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

import unittest
from argparse import ArgumentParser
from pathlib import Path
from unittest.mock import call, create_autospec, Mock, patch

import numpy as np
from PIL import Image

from vipersci.vis import anaglyph


class TestFunctions(unittest.TestCase):
    def setUp(self):
        self.left = np.array(
            [
                [0, 0, 0, 0],
                [0, 1, 0, 0],
                [0, 0, 1, 0],
                [0, 0, 0, 0],
            ]
        )
        self.right = np.array(
            [
                [0, 1, 0, 0],
                [0, 0, 1, 0],
                [0, 0, 0, 0],
                [0, 0, 0, 0],
            ]
        )

    def test_arg_parser(self):
        p = anaglyph.arg_parser()
        self.assertIsInstance(p, ArgumentParser)

    def test_correlate_and_shift(self):
        result = anaglyph.correlate_and_shift(self.left, self.right)
        np.testing.assert_array_equal(result, self.left)

    def test_anaglyph(self):
        truth = np.dstack((self.left, self.right, self.right))
        result = anaglyph.anaglyph(self.left, self.right)
        np.testing.assert_array_equal(result, truth)

        align_truth = np.dstack((self.left, self.left, self.left))
        align_result = anaglyph.anaglyph(self.left, self.right, align=True)
        np.testing.assert_array_equal(align_result, align_truth)

        rgb_result = anaglyph.anaglyph(
            np.dstack((self.left, self.left, self.left)),
            np.dstack((self.right, self.right, self.right)),
        )
        np.testing.assert_array_equal(rgb_result, truth)

    def test_anaglyph_exceptions(self):
        self.assertRaises(
            ValueError, anaglyph.anaglyph, self.left, np.delete(self.right, 3, 0)
        )
        self.assertRaises(
            ValueError, anaglyph.anaglyph, np.array([0, 1]), np.array([0, 1])
        )
        self.assertRaises(ValueError, anaglyph.anaglyph, self.left, np.array([0, 1]))

    @patch("vipersci.vis.anaglyph.imread")
    @patch("vipersci.vis.anaglyph.Image.fromarray")
    @patch("vipersci.vis.anaglyph.anaglyph")
    def test_create(self, m_anaglyph, m_fromarray, m_imread):
        m_im = create_autospec(Image)
        m_im.save = Mock()
        m_fromarray.return_value = m_im
        left = Path("dummy/left.tif")
        right = Path("dummy/right.tif")
        out = Path("out.tif")

        anaglyph.create(left, right, out)

        m_imread.assert_has_calls([call(left), call(right)])
        m_anaglyph.assert_called_once()
        m_fromarray.assert_called_once()
        m_im.save.assert_called_once_with(out)

    @patch("vipersci.vis.anaglyph.imread")
    @patch("vipersci.vis.anaglyph.Image.fromarray")
    def test_main(self, m_fromarray, m_imread):
        left = "left.tif"
        right = "right.tif"
        out = "out.tif"

        m_im = create_autospec(Image)
        m_im.save = Mock()
        m_fromarray.return_value = m_im

        pa_ret_val = anaglyph.arg_parser().parse_args(["-a", left, right, out])
        with patch("vipersci.vis.anaglyph.arg_parser") as parser:
            with patch("vipersci.vis.anaglyph.anaglyph") as m_anaglyph:
                parser.return_value.parse_args.return_value = pa_ret_val

                anaglyph.main()

                m_imread.assert_has_calls([call(Path(left)), call(Path(right))])
                m_anaglyph.assert_called_once()
                m_fromarray.assert_called_once()
                m_im.save.assert_called_once_with(Path(out))

        m_fromarray.reset_mock()
        m_imread.reset_mock()
        m_im.reset_mock()
        m_im.save.reset_mock()
        pa_ret_val = anaglyph.arg_parser().parse_args(["-s", left, right, out])
        with patch("vipersci.vis.anaglyph.arg_parser") as parser:
            parser.return_value.parse_args.return_value = pa_ret_val

            anaglyph.main()
            m_imread.assert_has_calls([call(Path(left)), call(Path(right))])
            m_fromarray.assert_called_once()
            m_im.save.assert_called_once_with(Path(out))

        m_fromarray.reset_mock()
        m_imread.reset_mock()
        m_im.reset_mock()
        m_im.save.reset_mock()
        pa_ret_val = anaglyph.arg_parser().parse_args(["-w", left, right, out])
        with patch("vipersci.vis.anaglyph.arg_parser") as parser:
            parser.return_value.parse_args.return_value = pa_ret_val

            anaglyph.main()
            m_imread.assert_has_calls([call(Path(left)), call(Path(right))])
            self.assertEqual(m_fromarray.call_count, 2)
            m_im.save.assert_called_once()

        m_fromarray.reset_mock()
        m_imread.reset_mock()
        m_im.reset_mock()
        m_im.save.reset_mock()
        pa_ret_val = anaglyph.arg_parser().parse_args(["-a", "-r", left, right, out])
        pa_ret_val.anaglyph = False
        with patch("vipersci.vis.anaglyph.arg_parser") as parser:
            with patch("vipersci.vis.anaglyph.correlate_and_shift") as m_corr_and_shift:
                parser.return_value.parse_args.return_value = pa_ret_val

                anaglyph.main()

                m_corr_and_shift.called_once()
                parser.error.called_once()



#!/usr/bin/env python
# coding: utf-8

"""Defines the VIS light records table using the SQLAlchemy ORM."""

# Copyright 2023, United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

from sqlalchemy import (
    Boolean,
    DateTime,
    Identity,
    Integer,
    String,
)
from sqlalchemy.orm import mapped_column, validates

from vipersci.vis.db import Base
import vipersci.vis.db.validators as vld


luminaire_names = {
    "navLeft": "NavLight Left",
    "navRight": "NavLight Right",
    "haz2": "HazLight Aft Port",
    "haz4": "HazLight Aft Starboard",
    "haz5": "HazLight Center Port",
    "haz6": "HazLight Center Starboard",
    "haz1": "HazLight Fore Port",
    "haz3": "HazLight Fore Starboard",
}

luminaire_shortnames = dict(
    hfp="haz1",
    hap="haz2",
    hfs="haz3",
    has="haz4",
    hcp="haz5",
    hcs="haz6",
    nl="navLeft",
    nr="navRight",
)


class LightRecord(Base):
    """An object to represent rows in the light_records table for VIS.  Each row
    represents a single state change (from 'on' to 'off' or vice-versa) for one
    light."""

    # This class is derived from SQLAlchemy's orm.DeclarativeBase
    # which means that it has a variety of class properties that are
    # then swept up into properties on the instantiated object via
    # super().__init__().

    # The table represents many of these objects, so the __tablename__ is
    # plural while the class name is singular.
    __tablename__ = "light_records"

    # The mapped_column() names below should use "snake_case" for the names that are
    # committed to the database as column names.  Furthermore, those names
    # should be similar, if not identical, to the PDS4 Class and Attribute
    # names that they represent.  Other names (like Yamcs parameter camelCase
    # names) are implemented as synonyms. Aside from the leading "id" column,
    # the remainder are in alphabetical order, since there are so many.

    id = mapped_column(Integer, Identity(start=1), primary_key=True)
    name = mapped_column(
        String, nullable=False, doc="The luminaire that was activated."
    )
    on = mapped_column(
        Boolean,
        nullable=False,
        doc="If this light, at this time, turned 'on', this column is TRUE, if the "
        "this light was detected to go from ON to OFF at this time, this column is "
        "FALSE.",
    )
    datetime = mapped_column(
        DateTime(timezone=True),
        nullable=False,
        doc="The time at which the luminaire was first measured to change state "
        "(via measuredState) from ON to OFF or vice-versa.",
    )

    @validates("name")
    def validate_name(self, key, value):
        if value not in luminaire_names.values():
            if value in luminaire_names:
                value = luminaire_names[value]
            else:
                raise ValueError(
                    f"The {key} ({value}) is not a luminaire name ({luminaire_names})"
                )
        return value

    @validates("datetime")
    def validate_datetime_asutc(self, key, value):
        return vld.validate_datetime_asutc(key, value)



"""Creates a panorama file and JSON stream suitable for use in MMGIS.
"""

# Copyright 2023, United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

import argparse
import json
import logging
from typing import Union, Optional
from pathlib import Path

import numpy as np
import numpy.typing as npt
from skimage.io import imread, imsave  # maybe just imageio here?
from skimage.exposure import equalize_adapthist, rescale_intensity
from skimage.transform import resize
from sqlalchemy import create_engine, select
from sqlalchemy.orm import Session

from vipersci.vis.db.pano_records import PanoRecord
from vipersci.pds import pid as pds
from vipersci import util

logger = logging.getLogger(__name__)

ImageType = Union[npt.NDArray[np.uint16], npt.NDArray[np.uint8]]


def arg_parser():
    parser = argparse.ArgumentParser(
        description=__doc__, parents=[util.parent_parser()]
    )
    parser.add_argument(
        "-d",
        "--dburl",
        help="Database with a raw_products table and a panorama_products table which "
        "will be read from and written to. If not given, no database will be "
        "written to. "
        "Example: postgresql://postgres:NotTheDefault@localhost/visdb",
    )
    parser.add_argument(
        "-m",
        "--mapserver",
        help="URL that will respond to requests when given an event_time and "
        "a crs_code.  Alternately, if a float is provided, this will be used as "
        "the rover yaw (zero==north) for testing.",
    )
    parser.add_argument(
        "-o",
        "--output_dir",
        type=Path,
        default=Path.cwd(),
        help="Output directory for label. Defaults to current working directory.  "
        "Output file names are fixed based on product_id, and will be over-written.",
    )
    parser.add_argument(
        "--prefix",
        type=Path,
        default=Path.cwd(),
        help="A directory path that, if given, will be prepended to paths given via "
        "inputs or will be prepended to the file_path values returned from a database "
        "query.",
    )
    parser.add_argument("input", help="Either VIS Pano product IDs or a JSON file.")
    return parser


def main():
    args = arg_parser().parse_args()
    util.set_logger(args.verbose)

    create_args = [args.input, args.prefix, args.output_dir, args.mapserver]

    if args.dburl is None:
        create_args.append(None)
        create(*create_args)
    else:
        engine = create_engine(args.dburl)
        with Session(engine) as session:
            create_args.append(session)
            create(*create_args)

            session.commit()

    return


def create(
    info: Union[Path, pds.PanoID, PanoRecord, str],
    prefixdir: Optional[Path] = None,
    outdir: Path = Path.cwd(),
    mapserver: Optional[str] = None,
    session: Optional[Session] = None,
    thumbsize=(None, 93),
):
    """
    Creates an MMGIS Panorama in *outdir*. Returns None.

    At this time, *input* should be a list of file paths or product IDs.

    If a path is provided to *outdir* the created files
    will be written there.

    If *session* is given, information for the PanoRecord will be
    written to the pano_records table.  If not, no database activity
    will occur.
    """

    # vid = None
    # pano = {}
    # source_path = Path()

    if isinstance(info, str):
        temp_vid = pds.PanoID(info)
        if str(temp_vid) == info:
            info = temp_vid

    if isinstance(info, pds.PanoID):
        if session is not None:
            pr = session.scalars(
                select(PanoRecord).where(PanoRecord.product_id == str(info))
            ).first()
            if pr is None:
                raise ValueError(f"{info} was not found in the database.")
            else:
                info = pr
        else:
            raise ValueError(f"Without a database session, can't lookup {info}")

    if isinstance(info, PanoRecord):
        # vid = pds.PanoID(info.product_id)
        source_path = (
            Path(info.file_path) if prefixdir is None else prefixdir / info.file_path
        )
        pano = info.asdict()
    elif isinstance(info, (Path, str)):
        # vid = pds.PanoID(info)
        with open(info) as f:
            pano = json.load(f)
        source_path = (
            Path(pano["file_path"])
            if prefixdir is None
            else prefixdir / pano["file_path"]
        )
    else:
        raise ValueError(
            f"an element in input is not the right type: {info} ({type(info)})"
        )

    if mapserver is None:
        yaw = 0.0
    else:
        try:
            yaw = float(mapserver)
        except ValueError:
            raise NotImplementedError("mapserver queries are not yet implemented.")

    # Convert to PNG
    image = equalize_adapthist(imread(str(source_path)))
    image8 = rescale_intensity(image, in_range="image", out_range="uint8").astype(
        "uint8"
    )

    outpath = outdir / Path(longname(source_path)).with_suffix(".png").name

    imsave(str(outpath), image8, check_contrast=False)

    d = mmgis_data(pano, yaw)
    # Decided not to include the URL key in the JSON this program outputs
    # as the process that takes these data and publishes them are what
    # should set that.
    # d["url"] = "not/sure/what/the/path/should/be/to/" + outpath.name

    # Make JPG thumbnail
    if thumbsize is not None:
        if isinstance(thumbsize, int):
            max_dim = max(np.shape(image8))
            if max_dim > thumbsize:
                scale = max_dim / thumbsize
                new_shape = tuple(int(x / scale) for x in np.shape(image8))
        elif len(thumbsize) == 2:
            if thumbsize[0] is not None:
                raise ValueError(
                    "Not sure how to handle a non-None first element for thumbsize."
                )
            scale = np.shape(image8)[1] / thumbsize[1]
            new_shape = tuple(int(x / scale) for x in np.shape(image8))

        image8 = rescale_intensity(
            resize(image8, new_shape), in_range="image", out_range="uint8"
        )
        outthumb = outpath.with_name(outpath.stem + "_thumb.jpeg")
        imsave(outthumb, image8, check_contrast=False)

    with open(outpath.stem + ".json", "w") as f:
        json.dump(d, f, indent=2, sort_keys=True)

    return


def longname(path):
    """Returns a longer iso8601-esque filename."""
    # YYYY-MM-DDTHH-mm-ss.SSS-pan
    vid = pds.PanoID(path.name)
    vdt = vid.datetime()
    return (
        vdt.date().isoformat() + "T" + vdt.time().isoformat().replace(":", "-") + "-pan"
    )


def mmgis_data(pano_data: dict, yaw=0.0):
    d = {
        "azmax": yaw + pano_data["rover_pan_max"],
        "azmin": yaw + pano_data["rover_pan_min"],
        "columns": pano_data["samples"],
        "elmax": pano_data["rover_tilt_max"],
        "elmin": pano_data["rover_tilt_min"],
        "isPanoramic": True,
        "name": pano_data["product_id"],
        "rows": pano_data["lines"],
    }
    return d



#!/usr/bin/env python
"""This module contains functions and data structures to enable
processing of PDS4 XML.
"""

# Copyright 2022, United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

im_version = "1.18.0.0"

dd = dict(
    pds="PDS4_PDS_1I00",
    disp="PDS4_DISP_1I00_1510",
    img="PDS4_IMG_1I00_1860",
    msn="PDS4_MSN_1I00_1300",
    proc="PDS4_PROC_1I00_1210",
)

ns = {}
for k in dd.keys():
    ns[k] = f"http://pds.nasa.gov/pds4/{k}/v1"


def find_text(root, xpath, unit_check=None, namespace=None):
    """Convenience function for returning the text from an element."""
    if namespace is None:
        namespace = ns

    element = root.find(xpath, namespace)
    if element is not None:
        if unit_check is not None:
            if element.get("unit") != unit_check:
                raise ValueError(
                    f"The {xpath} element does not have units of "
                    f"{unit_check}, has {element.get('unit')}"
                )
        el_text = element.text
        if el_text:
            return el_text
        else:
            raise ValueError(f"The XML {xpath} element contains no information.")
    else:
        raise ValueError(f"XML text does not have a {xpath} element.")



"""NIRVSS model module.

This module contains functions for simulating NIRVSS data.
"""

# Copyright 2022, United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

import numpy as np


def band_depth_H2O(temperature, ice_depth, a=0.001, b=100, c=50, d=1, e=10):
    """Returns the H2O band depth."""

    return (
        a
        * np.exp(-1 * np.absolute(temperature - b) / c)
        / np.exp(-1 * np.absolute(ice_depth - d) / e)
    )


def band_depth_OH(insolation, a=1e-3, b=150, c=0.5):
    """Returns the OH band depth."""
    return a * np.float_power(insolation - b, c)



"""Creates a PDS4 Bundle XML file from the provided Collection XML labels.
"""

# Copyright 2023, United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

import logging
from pathlib import Path
import xml.etree.ElementTree as ET

import pandas as pd
import yaml

from vipersci.pds.labelmaker import (
    assert_unique,
    gather_info,
    get_common_label_info,
    write_xml,
)
from vipersci.pds.xml import find_text, ns
from vipersci import util

logger = logging.getLogger(__name__)

# This definition is from Appendix K.2.1.6 (Bundle_Member_Entry) in the
# https://pds.nasa.gov/datastandards/documents/dph/current/PDS4_DataProvidersHandbook_1.20.0.pdf
collection_reference_type = {
    "Browse": "bundle_has_browse_collection",
    "Calibration": "bundle_has_calibration_collection",
    "Context": "bundle_has_context_collection",
    "Data": "bundle_has_data_collection",
    "Document": "bundle_has_document_collection",
    "Geometry": "bundle_has_geometry_collection",
    "Miscellaneous": "bundle_has_miscellaneous_collection",
    "SPICE Kernel": "bundle_has_spice_kernel_collection",
    "Schema": "bundle_has_schema_collection",
}


def add_parser(subparsers):
    parser = subparsers.add_parser("bundle", help=__doc__)

    parser.add_argument(
        "-c", "--config", type=Path, help="YAML file with configuration parameters."
    )
    parser.add_argument(
        "-t", "--template", type=Path, help="Genshi XML file template. "
    )
    parser.add_argument(
        "labelfiles",
        type=Path,
        nargs="+",
        help="Path(s) to all Collection XML label files to be included in the output "
        "Bundle label file.",
    )
    parser.set_defaults(func=main)
    return parser


def main(args):
    util.set_logger(args.verbose)

    d = yaml.safe_load(args.config.read_text())

    outpath = Path("bundle.xml")
    if outpath.exists():
        raise FileExistsError(f"The {outpath} file already exists.")

    labelinfo = []
    for labelpath in args.labelfiles:
        logger.info(f"Reading {labelpath}")
        labelinfo.append(get_label_info(labelpath))

    d.update(check_and_derive(d, labelinfo))

    write_xml(d, outpath, args.template)
    logger.info(f"Wrote {outpath}")

    return


def check_and_derive(config: dict, labelinfo: list):
    df = pd.DataFrame(labelinfo)

    # Check consistency for gathered labels:
    for x in (
        "bundle_lid",
        "investigation_name",
        "investigation_type",
        "investigation_lid",
        "host_name",
        "host_lid",
        "target_name",
        "target_type",
        "target_lid",
    ):
        assert_unique(config[x], df[x])

    # Generate values from gathered labels:
    d = gather_info(df, config["modification_details"])

    collections = []
    for label in labelinfo:
        collections.append(
            {
                "lid": label["lid"] + "::" + label["vid"],
                "type": collection_reference_type[label["collection_type"]],
            }
        )
    d["collections"] = collections

    return d


def get_label_info(path: Path) -> dict:
    root = ET.fromstring(path.read_text())

    if root.find("pds:Context_Area", ns) is not None:
        d = get_common_label_info(root, "pds:Context_Area")
    else:
        d = get_common_label_info(root, None)

    d["bundle_lid"] = ":".join(d["lid"].split(":")[:-1])
    d["collection_type"] = find_text(root, ".//pds:Collection/pds:collection_type")
    return d
